#include <assert.h>
#include <stdlib.h>
#include "clast_scanner.h"

void clast_scanner::print(clast_term *t)
{
    value_print(dst, VALUE_FMT, t->val);
    if (t->var)
	fprintf(dst, "*%s", t->var);
}

void clast_scanner::print(clast_binary *b)
{
    const char *s1, *s2;
    switch (b->type) {
	case clast_bin_mod:
	    s1 = "MOD(", s2 = ", ";
	    break;
	case clast_bin_div:
	    s1 = "(", s2 = ")/(";
	    break;
	case clast_bin_cdiv:
	    s1 = "DIVCEIL(", s2 = ", ";
	    break;
	case clast_bin_fdiv:
	    s1 = "DIVFLOOR(", s2 = ", ";
	    break;
	default:
	    assert(0);
    }
    fprintf(dst, s1);
    print(b->LHS);
    fprintf(dst, s2);
    value_print(dst, VALUE_FMT, b->RHS);
    fprintf(dst, ")");
}

void clast_scanner::print(clast_reduction *r)
{
    if (r->n == 1) {
	print(r->elts[0]);
	return;
    }

    const char *s1, *s2, *s3;

    switch (r->type) {
    case clast_red_sum:
	s1 = "", s2 = " + ", s3 = "";
	break;
    case clast_red_max:
	s1 = "MAX(", s2 = ", ", s3 = ")";
	break;
    case clast_red_min:
	s1 = "MIN(", s2 = ", ", s3 = ")";
	break;
    default:
	assert(0);
    }

    for (int i = 1; i < r->n; ++i)
	fprintf(dst, s1);
    print(r->elts[0]);
    for (int i = 1; i < r->n; ++i) {
	fprintf(dst, s2);
	print(r->elts[i]);
	fprintf(dst, s3);
    }
}

void clast_scanner::print(clast_expr *e)
{
    switch (e->type) {
    case clast_expr::expr_term:
	print((clast_term*) e);
	break;
    case clast_expr::expr_red:
	print((clast_reduction*) e);
	break;
    case clast_expr::expr_bin:
	print((clast_binary*) e);
	break;
    default:
	assert(0);
    }
}

void clast_scanner::print(clast_assignment *a)
{
    print_indent();
    fprintf(dst, "%s = ", a->LHS);
    print(a->RHS);
    fprintf(dst, ";\n", a->LHS);
}

void clast_scanner::print(clast_guard *g)
{
    int n = g->n;
    print_indent();
    fprintf(dst, "if (");
    for (int i = 0; i < n; ++i) {
	if (i > 0)
	    fprintf(dst," && ");
	fprintf(dst,"(");
	print(g->eq[i].LHS);
	if (g->eq[i].sign == 0)
	    fprintf(dst," == ");
	else if (g->eq[i].sign > 0)
	    fprintf(dst," >= ");
	else
	    fprintf(dst," <= ");
	print(g->eq[i].RHS);
	fprintf(dst,")");
    }
    fprintf(dst, ") {\n");
    indent += 4;
    print(g->then);
    indent -= 4;
    print_indent();
    fprintf(dst, "}\n");
}

void clast_scanner::print(clast_for *f)
{
    assert(f->LB && f->UB);
    print_indent();
    fprintf(dst, "for (%s=", f->iterator);
    print(f->LB);
    fprintf(dst, "; %s<=", f->iterator);
    print(f->UB);
    fprintf(dst, "; ++%s) {\n", f->iterator);
    indent += 4;
    print(f->body);
    indent -= 4;
    print_indent();
    fprintf(dst, "}\n");
}

void clast_scanner::print(clast_stmt *s)
{
    for ( ; s; s = s->next) {
	if (CLAST_STMT_IS_A(s, stmt_root))
	    continue;
	if (CLAST_STMT_IS_A(s, stmt_ass)) {
	    print((clast_assignment *) s);
	} else if (CLAST_STMT_IS_A(s, stmt_user)) {
	    print_user((clast_user_stmt *) s);
	} else if (CLAST_STMT_IS_A(s, stmt_for)) {
	    print((clast_for *) s);
	} else if (CLAST_STMT_IS_A(s, stmt_guard)) {
	    print((clast_guard *) s);
	} else {
	    assert(0);
	}
    }
}

void clast_scanner::print_header()
{
    fprintf(dst, "%s", "#include <assert.h>\n");
    fprintf(dst, "%s", "#define MIN(x,y)\t" "((x)<(y)?(x):(y))" "\n");
    fprintf(dst, "%s", "#define MAX(x,y)\t" "((x)>(y)?(x):(y))" "\n");
    fprintf(dst, "%s", "#define MOD(x,y)\t" "((((x) % (y)) < 0) ? "
	"(((y) < 0) ? (((x) % (y)) - (y)) : (((x) % (y)) + (y))) : ((x) % (y)))"
	"\n");
    fprintf(dst, "%s", "#define DIVCEIL(x,y)\t" "(((x)<0)^((y)<0) ? (x)/(y) : "
	"((x) < 0 ? ((x)+(y)+1)/(y) : ((x)+(y)-1)/(y)))" "\n");
    fprintf(dst, "%s", "#define DIVFLOOR(x,y)\t"
	"(((x)<0)^((y)<0) ? ((x) < 0 ? ((x)-(y)+1)/(y) : ((x)-(y)-1)/(y)) : "
	"(x)/(y))" "\n");
}

void clast_scanner::print_iterators(CloogNames* names)
{
    print_indent();
    if (names->nb_iterators <= 0)
	return;
    fprintf(dst, "int ");
    for (int i = 0; i < names->nb_iterators; ++i) {
	if (i)
	    fprintf(dst, ", ");
	fprintf(dst, names->iterators[i]);
    }
    fprintf(dst, ";\n");
}
