#include "suif_inc.h"
#include <map>
#include <sstream>
#include "dump.h"
#include "pdg.h"
#include "pdgutil.h"

#ifndef HAVE_STRNDUP
extern "C" char *strndup(const char *s, size_t n);
#endif

using pdg::PDG;
using pdg::array;
using pdg::OUnionSet;

void
dumper::visit(tree_node *t)
{
    if (t->peek_annote(k_static_root)) {
	/* Only dump a single domain */
	if (root == 0)
	    root = t->number();
	else
	    root = -1;
    }

    if (root <= 0)
	return;

    if (t->kind() != TREE_INSTR)
	return;
    instruction *in = ((tree_instr*)t)->instr();
    domain_ann *d = (domain_ann *)in->peek_annote(k_iteration_domain);
    if (d == NULL)
	return;

    nodes.push_back(d);
}

dumper::~dumper()
{
}

string c_type(type_node *tn)
{
    std::ostringstream strm;

    switch (tn->op()) {
    case TYPE_INT: {
        base_type *bt = (base_type *) tn;

        if (!bt->is_signed())
	    strm << "unsigned ";
	switch (c_int_type(tn)) {
	case C_char:
	    if (bt->is_signed())
		strm << "signed ";
	    strm << "char";
	    break;
	case C_int:
	    strm << "int";
	    break;
	case C_short:
	    strm << "short";
	    break;
	case C_long:
	    strm << "long";
	    break;
	case C_longlong:
	    strm << "long long";
	    break;
	default:
	    assert(0);
	}
	break;
    }
    case TYPE_FLOAT: {
	switch (c_float_type(tn)) {
	case C_float:
	    strm << "float";
	    break;
	case C_double:
	    strm << "double";
	    break;
	case C_longdouble:
	    strm << "long double";
	    break;
	default:
	    assert(0);
	}
	break;
    }
    case TYPE_STRUCT: {
        struct_type *st = (struct_type *) tn;
	strm << "struct " << st->name();
	break;
    }
    default:
	assert(0);
    }

    return strm.str();
}

static int compute_total_read(struct stmt_node *node, int c)
{
	if (node->is_access()) {
		stmt_access *sa = static_cast<stmt_access *>(node);
		return c + !sa->write;
	}

	stmt_call *sc = static_cast<stmt_call *>(node);

	for (int i = 0; i < sc->arg.size(); ++i)
		c = compute_total_read(sc->arg[i], c);

	return c;
}

static int compute_total_read(struct stmt *s)
{
	return compute_total_read(s->tree, 0);
}

pdg::access *dumper::add_access(pdg::node *pn, stmt_access *sa)
{
	var_sym *av = sa->array;
	if (av && arrays.find(av) == arrays.end()) {
		array *a = new array();
		a->id = av->sym_id();
		a->name = new str(av->name());
		type_node *tn = av->type();
		while (tn->is_array()) {
			array_type *an = dynamic_cast<array_type*>(tn);
			assert(an->lower_bound().is_constant());
			assert(an->upper_bound().is_constant());
			tn = an->elem_type();
			a->dims.push_back(an->upper_bound().constant() -
					  an->lower_bound().constant() + 1);
		}
		a->element_type = new str(c_type(tn));
		pdg->arrays.push_back(a);
		arrays[av] = a;
	}
	pdg::access *a = new pdg::access;
	a->array = av ? arrays[av] : NULL;
	bool write = sa->write;
	a->type = write ? pdg::access::write : pdg::access::read;
	a->map = new SMap(pn->source->dim, sa->output, sa->nlq, globals_sym);
	a->map->params = pdg->params;
	if (write)
		a->nr = totalread + nwrite++;
	else
		a->nr = nread++;
	pn->statement->accesses.push_back(a);
	return a;
}

pdg::call_or_access *dumper::scan_stmt(pdg::node *pn, stmt_node *n)
{
	pdg::call_or_access *coa = new pdg::call_or_access;

	if (n->is_access()) {
		coa->type = pdg::call_or_access::t_access;
		coa->access = add_access(pn, static_cast<stmt_access *>(n));
	} else {
		stmt_call *sc = static_cast<stmt_call *>(n);
		coa->type = pdg::call_or_access::t_call;
		coa->call = new pdg::function_call;
		coa->call->name = new str(string(sc->name));

		for (int i = 0; i < sc->arg.size(); ++i) {
			pdg::call_or_access *child;
			child = scan_stmt(pn, sc->arg[i]);
			coa->call->arguments.push_back(child);
		}
	}

	return coa;
}

/* Convert stmt to pdg::function_call and add all accesses to ps */
pdg::function_call *dumper::scan_stmt(pdg::node *pn, stmt *s)
{
	totalread = compute_total_read(s);
	nread = 0;
	nwrite = 0;

	pdg::function_call *call;
	pdg::call_or_access *coa = scan_stmt(pn, s->tree);

	for (int i = 0; i < s->out.size(); ++i)
		add_access(pn, s->out[i]);

	if (coa->type == pdg::call_or_access::t_access) {
		call = new pdg::function_call;
		call->arguments.push_back(coa);
		call->name = new str(string(""));
	} else {
		call = coa->call;
		delete coa;
	}
	return call;
}

void dumper::dump(int argc, char * argv[])
{
    if (root <= 0)
	return;

    pdg = new PDG;
    pdg->placement = new str(string("original"));
    pdg->root = root;

    const char *name = tp->proc()->file()->name();
    char *dot = strrchr(name, '.');
    if (!dot)
	pdg->name = new str(string(name));
    else {
	char *basename = strndup(name, dot-name);
	pdg->name = new str(string(basename));
	free(basename);
    }

    std::vector<Global_Var_Decl *> globals;
    map<var_sym *, Param_Iterator *>::iterator gi;
    for (gi = sa->vars.begin(); gi != sa->vars.end(); ++gi) {
	if ((*gi).second->iterator)
	    continue;
	globals.push_back((*gi).second);
	globals_sym.push_back((*gi).first);
	pdg::parameter *p = new pdg::parameter;
	p->id = (*gi).first->sym_id();
	p->name = new str((string)(*gi).second->base_name());

	file_symtab *symtab = tp->proc()->file()->symtab();
	var_def_list_iter j(symtab->var_defs());
	while (!j.is_empty()) {
	    var_def *def = j.step();
	    if (def->variable() != (*gi).first)
		continue;
	    annote_list_iter i(def->annotes());
	    while (!i.is_empty()) {
		annote *an = i.step();
		if (an->name() != k_repeat_init)
		    continue;
		immed_list *ims = an->immeds();
		if ((*ims)[2].kind() != im_int)
		    continue;
		p->value = new integer((*ims)[2].integer());
	    }
	}

	pdg->params.push_back(p);
    }

    if (ctx) {
	pdg->context = new OUnionSet(ctx->ctx, globals);
	pdg->context->params = pdg->params;
    }

    for (int i = 0; i < nodes.size(); ++i) {
	domain_ann *n = nodes[i];
	if (!n->domain->is_satisfiable())
	    continue;
	pdg::node * pn = new pdg::node;
	pdg::statement * ps = new pdg::statement;
	pdg->nodes.push_back(pn);
	pn->nr = i;
	pn->statement = ps;
	ps->operation = n->insn->number();
	ps->line = source_line_num(n->insn);
	pn->source = new OUnionSet(n->domain, globals);
	pn->source->params = pdg->params;
	for (int j = 0; j < n->plevel; ++j)
	    pn->prefix.push_back(n->prefix[j]);
	ps->top_function = scan_stmt(pn, n->statement);
    }

    pdg->add_history_line("pers", argc, argv);
    pdg->Dump();
    pdg->free();
    delete pdg;
}

static void
visit (tree_node *t, void *d)
{
    ((dumper *)d)->visit(t);
}

void dumper::scan()
{
    root = 0;
    tp->map(::visit, (void *) this);
}

void dump(tree_proc *tp, context_info *ctx, int argc, char * argv[])
{
    dumper d(tp, ctx);
    d.scan();
    d.dump(argc, argv);
}
