#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <algorithm>
#include <map>
#include <set>
#include <vector>
#include "argp.h"
#include "progname.h"

#include "yaml.h"
#include "pdg.h"

#include "isl_set.h"
#include "isl_map.h"

static const char args_doc[] = "program1 program2";

struct arguments {
	const char *program[2];

	std::vector<const char *> commutative;
};

struct argp_option argp_options[] = {
	{ "commutative",	'c',	"operators", 0,
		"comma separated list of commutative operators" },
	{ 0 }
};

error_t parse_opt(int key, char *arg, struct argp_state *state)
{
	char *tok;
	struct arguments *arguments = (struct arguments *)(state->input);

	switch (key) {
	case ARGP_KEY_ARG:
		if (state->arg_num >= 2)
			argp_usage (state);
		arguments->program[state->arg_num] = arg;
		break;
	case ARGP_KEY_END:
		if (state->arg_num < 2)
			argp_usage (state);
		break;
	case 'c':
		tok = strtok(arg, ",");
		if (!tok)
			break;
		arguments->commutative.push_back(strdup(tok));
		while ((tok = strtok(NULL, ",")) != NULL)
			arguments->commutative.push_back(strdup(tok));
		break;
	default:
		return ARGP_ERR_UNKNOWN;
	}

	return 0;
}

struct computation;

struct edge {
	computation		*source;
	isl_ctx			*ctx;
	isl_basic_map		*relation;
	unsigned		 pos;

	edge(isl_ctx *ctx) : ctx(ctx) {}
	~edge() {
		isl_basic_map_free(ctx, relation);
	}

	void dump(isl_ctx *ctx);
};

struct computation {
	isl_ctx			*ctx;
	isl_set			*domain;
	const char		*operation;
	unsigned		 arity;
	int			 location;
	unsigned		 input : 1;

	std::vector<edge *>	 edges;

	computation(isl_ctx *ctx) : ctx(ctx), input(0) {}

	~computation() {
		for (int i = 0; i < edges.size(); ++i)
			delete edges[i];
		isl_set_free(ctx, domain);
		free((char *)operation);
	}

	bool is_input() const {
		return input;
	}

	bool is_copy() const {
		return !input && !strcmp(operation, "");
	}

	void dump(isl_ctx *ctx);
};

struct input_array : computation {
	input_array(isl_ctx *ctx, const char *name) : computation(ctx) {
		domain = NULL;
		operation = strdup(name);
		arity = 0;
		location = -1;
		input = 1;
	}
};

struct dependence_graph {
	std::vector<char *>	 	 output_arrays;
	std::vector<computation *>	 vertices;
	computation 			*out;

	dependence_graph() : out(NULL) {}
	~dependence_graph() {
		for (int i = 0; i < output_arrays.size(); ++i)
			free(output_arrays[i]);
		for (int i = 0; i < vertices.size(); ++i)
			delete vertices[i];
		delete out;
	}

	void dump(isl_ctx *ctx);
};

void edge::dump(isl_ctx *ctx)
{
	fprintf(stderr, "\tpos: %d, source: %d,%s [%p]\n",
			pos, source->location, source->operation, source);
	isl_basic_map_dump(ctx, relation, stderr, 8);
}

void computation::dump(isl_ctx *ctx)
{
	fprintf(stderr, "op: %d,%s [%p]\n", location, operation, this);
	isl_set_dump(ctx, domain, stderr, 0);

	for (int i = 0; i < edges.size(); ++i) {
		fprintf(stderr, "edge %d:\n", i);
		edges[i]->dump(ctx);
	}
}

void dependence_graph::dump(isl_ctx *ctx)
{
	for (int i = 0; i < vertices.size(); ++i)
		vertices[i]->dump(ctx);
	out->dump(ctx);
}

static void add_input_edges(isl_ctx *ctx, int pos, pdg::dependence *dep,
			computation *comp, computation *source)
{
	pdg::access *a = dep->to_access;
	assert(source);
	isl_map *map = a->map->get_isl_map(ctx);
	isl_map *relation = dep->relation->get_isl_map(ctx);
	isl_set *domain = isl_map_range(ctx, relation);
	map = isl_map_intersect_domain(ctx, map, domain);
	for (int i = 0; i < map->n; ++i) {
		edge *e = new edge(ctx);
		e->source = source;
		e->pos = pos;
		e->relation = isl_basic_map_copy(ctx, map->p[i]);
		comp->edges.push_back(e);
	}
	isl_map_free(ctx, map);
}

/* For each (flow) dependence with write access a, and an
 * edge to computation comp.
 * Note that in the pdg, the dependences have as "input" the source (read)
 * and as "output" the sink (write), while in the dependence_graph,
 * we want edges that point from sink to source.
 */
static void add_dep_edges(isl_ctx *ctx, pdg::PDG *pdg,
			    std::map<pdg::node *, computation *> &n2c,
			    std::map<pdg::array *, computation *> &a2c,
			    int pos,
			    pdg::access *a, computation *comp)
{
	for (int i = 0; i < pdg->dependences.size(); ++i) {
		pdg::dependence *dep = pdg->dependences[i];
		if (dep->to_access != a)
			continue;

		if (dep->type == pdg::dependence::uninitialized) {
			add_input_edges(ctx, pos, dep, comp, a2c[dep->array]);
			continue;
		}
		assert(dep->type == pdg::dependence::flow);

		isl_map *relation = dep->relation->get_isl_map(ctx);
		relation = isl_map_reverse(ctx, relation);

		for (int j = 0; j < relation->n; ++j) {
			edge *e = new edge(ctx);
			e->source = n2c[dep->from];
			assert(e->source);
			e->pos = pos;
			e->relation = isl_basic_map_copy(ctx, relation->p[j]);
			comp->edges.push_back(e);
		}
		isl_map_free(ctx, relation);
	}
}

/* For each argument of function call call, we add one or more edges to
 * computation comp.
 * In case the argument is an access, the edges corresponding to
 * the dependences are added in add_dep_edges.
 * If it is itself a function call, then a single edge is added
 * to a new child and add_edges is called recursively on the
 * nested function call.
 */
static void add_edges(isl_ctx *ctx, dependence_graph *dg, pdg::PDG *pdg,
			std::map<pdg::node *, computation *> &n2c,
			std::map<pdg::array *, computation *> &a2c,
			pdg::function_call *call, computation *comp)
{
	unsigned nparam = pdg->params.size();

	for (int i = 0; i < call->arguments.size(); ++i) {
		pdg::call_or_access *coa = call->arguments[i];

		if (coa->type == pdg::call_or_access::t_access) {
			add_dep_edges(ctx, pdg, n2c, a2c, i, coa->access, comp);
		} else {
			computation *child = new computation(ctx);
			child->domain = isl_set_copy(ctx, comp->domain);
			child->operation = strdup(coa->call->name->s.c_str());
			child->arity = coa->call->arguments.size();
			dg->vertices.push_back(child);

			isl_assert(ctx, comp->domain->n > 0, goto error);
			unsigned dim = comp->domain->p[0]->dim;

			edge *e = new edge(ctx);
			e->source = child;
			e->relation = isl_basic_map_identity(ctx, nparam, dim);
			e->pos = i;
			comp->edges.push_back(e);

			add_edges(ctx, dg, pdg, n2c, a2c, coa->call, child);
		}
	}

	return;
error:
	abort();
}

/* Add an edge from the output computation (out) to a computation computing
 * (part of) an output array.  node is pdg node corresponding to this
 * computation.  It is assumed to have a single write access, corresponding
 * to the output array.  exposed is that part of the node domain that
 * is not the source of an output dependence, i.e., those for which
 * the computed array elements are not overwritten.
 * Let f(i) be the index expression of the write access and let
 * pos be the position of the output array in the sorted list of
 * output arrays, then an edge is added to out with relation
 *
 *	{ [ f(i), 0, pos ] -> [i] | i \in exposed }
 *
 * where the total dimension of the domain is out_dim, and this domain
 * is added to the domain of the output computation.
 */
static void add_output_edge(isl_ctx *ctx, computation *out, pdg::node *node,
	std::map<pdg::array *, int> &array2pos,
	std::map<pdg::node *, computation *> &n2c,
	isl_set *exposed, unsigned nparam, unsigned out_dim)
{
	pdg::statement *s = node->statement;
	int j;
	for (j = 0;  j < s->accesses.size(); ++j)
		if (s->accesses[j]->type == pdg::access::write &&
		    s->accesses[j]->array->type == pdg::array::output)
			break;
	pdg::array *array = s->accesses[j]->array;
	int pos = array2pos[array];
	isl_map *rel = isl_map_reverse(ctx,
			s->accesses[j]->map->get_isl_map(ctx));
	rel = isl_map_extend(ctx, rel, nparam, out_dim, rel->n_out);
	rel = isl_map_intersect_range(ctx, rel, exposed);
	rel = isl_map_fix_input_si(ctx, rel, out_dim-1, pos);
	for (int k = array->dims.size(); k < out_dim-1; ++k)
		rel = isl_map_fix_input_si(ctx, rel, k, 0);

	out->domain = isl_set_union_disjoint(ctx, out->domain,
				isl_map_domain(ctx,
					isl_map_copy(ctx, rel)));
	for (int j = 0; j < rel->n; ++j) {
		edge *e = new edge(ctx);
		e->source = n2c[node];
		assert(e->source);
		e->pos = 0;
		e->relation = isl_basic_map_copy(ctx, rel->p[j]);
		out->edges.push_back(e);
	}
	isl_map_free(ctx, rel);
}

static void ensure_output_deps(pdg::PDG *pdg, pdg::array *array)
{
	for (int i = 0; i < array->analysis_performed.size(); ++i)
		if (array->analysis_performed[i]->type ==
						    pdg::dependence::output)
			return;
	fprintf(stderr, "no output dependence analysis performed on array %s\n",
		array->name->s.c_str());
	exit(-1);
}

struct less_name : public std::binary_function<pdg::array*, pdg::array*, bool> {
	bool operator()(pdg::array* x, pdg::array* y) {
		return strcmp(x->name->s.c_str(), y->name->s.c_str()) < 0;
	}
};

/* Convert a pdg to a dependence_graph containing all the information
 * needed for equivalence checking.
 *
 * We first compute the total dimension of the output computation
 * and create input computations for each input array.
 * Then we create computations for each node in the pdg and
 * a special output computation.
 * For each node, we subsequently compute all the edges emanating
 * from the corresponding computation.
 * Finally, for each node writing to an output array, we compute
 * that part of the domain that computes part of the final
 * contents of the output array and add corresponding edges
 * to the output computation.
 */
static dependence_graph *pdg_to_dg(isl_ctx *ctx, pdg::PDG *pdg)
{
	unsigned nparam = pdg->params.size();
	dependence_graph *dg = new dependence_graph;
	std::map<pdg::node *, computation *> n2c;
	std::map<pdg::array *, computation *> a2c;

	unsigned out_dim = 0;
	std::vector<pdg::array *> output_arrays;
	std::map<pdg::array *, int> array2pos;
	std::map<pdg::node *, isl_set *> exposed;

	for (int i = 0; i < pdg->arrays.size(); ++i) {
		pdg::array *array = pdg->arrays[i];
		if (array->type == pdg::array::output) {
			if (array->dims.size() + 1 > out_dim)
				out_dim = array->dims.size() + 1;
			output_arrays.push_back(array);
			ensure_output_deps(pdg, array);
		}

		if (array->type != pdg::array::input)
			continue;

		computation *comp;
		comp = new input_array(ctx, array->name->s.c_str());
		a2c[array] = comp;
		dg->vertices.push_back(comp);
	}
	if (!out_dim) {
		fprintf(stderr, "no output arrays\n");
		exit(-1);
	}
	for (int i = 0; i < pdg->nodes.size(); ++i) {
		pdg::node *node = pdg->nodes[i];
		pdg::statement *s = node->statement;

		computation *comp = new computation(ctx);
		n2c[node] = comp;
		comp->domain = node->source->get_isl_set(ctx);
		comp->operation = strdup(s->top_function->name->s.c_str());
		comp->arity = s->top_function->arguments.size();
		comp->location = s->line;
		dg->vertices.push_back(comp);
	}
	std::sort(output_arrays.begin(), output_arrays.end(), less_name());
	for (int i = 0; i < output_arrays.size(); ++i) {
		dg->output_arrays.push_back(strdup(output_arrays[i]->name->s.c_str()));
		array2pos[output_arrays[i]] = i;
	}
	dg->out = new computation(ctx);
	dg->out->domain = isl_set_empty(ctx, nparam, out_dim);
	dg->out->operation = strdup("#Out");
	dg->out->arity = 1;
	dg->out->location = -1;
	for (int i = 0; i < pdg->nodes.size(); ++i) {
		pdg::node *node = pdg->nodes[i];
		pdg::statement *s = node->statement;
		computation *comp = n2c[node];

		add_edges(ctx, dg, pdg, n2c, a2c, s->top_function, comp);

		int j;
		for (j = 0;  j < s->accesses.size(); ++j)
			if (s->accesses[j]->type == pdg::access::write &&
			    s->accesses[j]->array->type == pdg::array::output)
				break;
		if (j >= s->accesses.size())
			continue;
		exposed[node] = node->source->get_isl_set(ctx);
	}
	for (int i = 0; i < pdg->dependences.size(); ++i) {
		pdg::dependence *dep = pdg->dependences[i];
		if (dep->type != pdg::dependence::output)
			continue;
		pdg::node *node = dep->from;
		if (!exposed[node] || isl_set_is_empty(ctx, exposed[node]))
			continue;
		isl_map *rel = dep->relation->get_isl_map(ctx);
		exposed[node] = isl_set_subtract(ctx, exposed[node],
					isl_map_domain(ctx, rel));
	}
	for (int i = 0; i < pdg->nodes.size(); ++i) {
		pdg::node *node = pdg->nodes[i];
		if (!exposed[node] || isl_set_is_empty(ctx, exposed[node])) {
			isl_set_free(ctx, exposed[node]);
			continue;
		}
		add_output_edge(ctx, dg->out, node, array2pos, n2c,
				exposed[node], nparam, out_dim);
	}
	return dg;
error:
	delete dg;
	return NULL;
}

struct eq_node {
	computation		*comp[2];
	isl_ctx			*ctx;
	isl_basic_map		*want;
	isl_map			*need;
	isl_map			*got;
	std::set<eq_node *>	 assumed;
	unsigned		 closed : 1;
	unsigned		 narrowing : 1;
	unsigned		 widening : 1;
	unsigned		 invalidated : 1;
	unsigned		 reset : 1;

	eq_node(isl_ctx *ctx, computation *c1, computation *c2,
			isl_basic_map *w) : ctx(ctx), want(w), closed(0),
			widening(0), narrowing(0), invalidated(0), reset(0),
			need(NULL), got(NULL) {
		comp[0] = c1;
		comp[1] = c2;
	}
	bool is_still_valid();
	void collect_open_assumed(std::set<eq_node *> &c);
	~eq_node() {
		isl_basic_map_free(ctx, want);
		isl_map_free(ctx, need);
		isl_map_free(ctx, got);
	}
};

bool eq_node::is_still_valid()
{
	if (invalidated)
		return 0;

	std::set<eq_node *>::iterator i;
	for (i = assumed.begin(); i != assumed.end(); ++i) {
		assert(*i != this);
		if ((*i)->reset ||
		    ((*i)->closed && !(*i)->is_still_valid())) {
			invalidated = 1;
			return 0;
		}
	}
	return 1;
}

void eq_node::collect_open_assumed(std::set<eq_node *> &c)
{
	std::set<eq_node *>::iterator i;
	for (i = assumed.begin(); i != assumed.end(); ++i) {
		if ((*i)->closed)
			(*i)->collect_open_assumed(c);
		else
			c.insert(*i);
	}
}

/* A comp_pair contains all the edges that have the same pair
 * of computations.
 */
struct comp_pair {
	computation		*comp[2];

	std::vector<eq_node *>	 nodes;

	eq_node *tabled(eq_node *node);
	eq_node *last_ancestor(eq_node *node);
	~comp_pair();
};

comp_pair::~comp_pair()
{
	std::vector<eq_node *>::iterator i;
	for (i = nodes.begin(); i != nodes.end(); ++i)
		delete *i;
}

eq_node *comp_pair::tabled(eq_node *node)
{
	std::vector<eq_node *>::iterator i;

	for (i = nodes.begin(); i != nodes.end(); ++i) {
		if (*i == node)
			continue;
		if (!(*i)->closed)
			continue;
		if (!(*i)->is_still_valid())
			continue;
		int is_subset;
		is_subset = isl_basic_map_is_subset(node->ctx,
				node->want, (*i)->want);
		assert(is_subset >= 0);
		if (!is_subset)
			continue;
		return *i;
	}
	return NULL;
}

eq_node *comp_pair::last_ancestor(eq_node *node)
{
	std::vector<eq_node *>::reverse_iterator i;
	for (i = nodes.rbegin(); i != nodes.rend(); ++i) {
		if (*i == node)
			continue;
		if ((*i)->closed)
			continue;
		return *i;
	}
	return NULL;
}

typedef std::pair<computation *, computation *> computation_pair;
typedef std::map<computation_pair, comp_pair *> c2p_t;

struct equivalence_checker {
	c2p_t c2p;
	arguments *arg;

	equivalence_checker(arguments *a) : arg(a) {}
	comp_pair *get_comp_pair(eq_node *node);
	void handle(eq_node *node);
	void dismiss(eq_node *node);
	void handle_propagation(eq_node *node);
	void handle_copy(eq_node *node, int i);
	void handle_with_ancestor(eq_node *node, eq_node *a);
	void handle_propagation_same(eq_node *node);
	void handle_narrowing(eq_node *node);

	bool is_commutative(const char *op);

	~equivalence_checker();
};

equivalence_checker::~equivalence_checker()
{
	c2p_t::iterator i;

	for (i = c2p.begin(); i != c2p.end(); ++i)
		delete (*i).second;
}

bool equivalence_checker::is_commutative(const char *op)
{
	std::vector<const char *>::iterator iter;

	for (iter = arg->commutative.begin();
	     iter != arg->commutative.end(); ++iter)
		if (!strcmp(*iter, op))
			return 1;
	return 0;
}

comp_pair *equivalence_checker::get_comp_pair(eq_node *node)
{
	c2p_t::iterator i;
	comp_pair *cp;

	i = c2p.find(computation_pair(node->comp[0], node->comp[1]));
	if (i == c2p.end()) {
		cp = new comp_pair;
		c2p[computation_pair(node->comp[0], node->comp[1])] = cp;
	} else
		cp = (*i).second;
	return cp;
}

void equivalence_checker::dismiss(eq_node *node)
{
	/* unclosed nodes weren't added to c2p->nodes */
	if (node && !node->closed)
		delete node;
}

/* Check for which subset (got) of the want relation equivelance
 * holds for the pair of computations in the equivalence node.
 *
 * We first handle the easy cases: empty want, input computations
 * and tabled nodes.
 *
 * If the current node is not a narrowing or a widening node
 * and we can find an ancestor with the same pair of computations,
 * then we will try to apply induction or widening in handle_with_ancestor.
 * However, if the requested relation (want) of the ancestor is a strict
 * subset of that of the current node, then we have already applied
 * widening on an intermediate node (with a different pair of computations)
 * so we shouldn't apply widening again (and we can't apply induction
 * because the relation of the ancestor is a strict subset).
 *
 * In all other cases we try to apply propagation in handle_propagation.
 */
void equivalence_checker::handle(eq_node *node)
{
	eq_node *s = NULL;
	eq_node *a = NULL;

	isl_basic_map *want = node->want;
	node->need = isl_map_empty(node->ctx,
			want->nparam, want->n_in, want->n_out);
	int empty = isl_basic_map_is_empty(node->ctx, want);
	assert(empty >= 0);
	if (empty) {
		node->got = isl_map_empty(node->ctx,
				want->nparam, want->n_in, want->n_out);
		return;
	}
	if (node->comp[0]->is_input() && node->comp[1]->is_input()) {
		if (strcmp(node->comp[0]->operation, node->comp[1]->operation)) {
			node->got = isl_map_empty(node->ctx,
					want->nparam, want->n_in, want->n_out);
			return;
		}
		node->got = isl_map_from_basic_map(node->ctx,
			    isl_basic_map_intersect(node->ctx,
				isl_basic_map_copy(node->ctx, node->want),
				isl_basic_map_identity(node->ctx,
					node->want->nparam, node->want->n_in)));
		return;
	}

	comp_pair *cp = get_comp_pair(node);
	if ((s = cp->tabled(node)) != NULL) {
		node->got = isl_map_intersect(node->ctx,
				isl_map_copy(node->ctx, s->got),
				isl_map_from_basic_map(node->ctx,
				isl_basic_map_copy(node->ctx, node->want)));
		s->collect_open_assumed(node->assumed);
		return;
	}

	cp->nodes.push_back(node);

	if (!node->narrowing && !node->widening &&
		    (a = cp->last_ancestor(node)) != NULL) {
		int is_subset;
		is_subset = isl_basic_map_is_strict_subset(node->ctx,
							a->want, node->want);
		assert(is_subset >= 0);
		if (is_subset)
			handle_propagation(node);
		else
			handle_with_ancestor(node, a);
	} else
		handle_propagation(node);
	node->closed = 1;
}

/* Check if we can apply propagation to prove equivalence of the given node.
 * First check if we need to apply copy propagation and if not
 * check if the operations are the same and apply "regular" propagation.
 *
 * After we get back, we need to check that any induction hypotheses
 * we have used in the process of proving the node hold.
 * If not, we replace the obtained relation (got) by that
 * of a narrowing node in handle_narrowin.
 */
void equivalence_checker::handle_propagation(eq_node *node)
{
	if (node->comp[0]->is_copy()) {
		handle_copy(node, 0);
	} else if (node->comp[1]->is_copy()) {
		handle_copy(node, 1);
	} else if (!strcmp(node->comp[0]->operation, node->comp[1]->operation) &&
			node->comp[0]->arity == node->comp[1]->arity) {
		handle_propagation_same(node);
	} else {
		isl_basic_map *want = node->want;
		node->got = isl_map_empty(node->ctx,
				want->nparam, want->n_in, want->n_out);
		return;
	}
	node->assumed.erase(node);
	int is_subset = isl_map_is_subset(node->ctx, node->need, node->got);
	assert(is_subset >= 0);
	if (!is_subset)
		handle_narrowing(node);
}

static eq_node *propagation_node(eq_node *node, edge *e1, edge *e2)
{
	isl_basic_map *new_want;
	computation *comp1 = e1 ? e1->source : node->comp[0];
	computation *comp2 = e2 ? e2->source : node->comp[1];

	if (!comp1->is_copy() && !comp2->is_copy() &&
	    strcmp(comp1->operation, comp2->operation))
		return NULL;

	new_want = isl_basic_map_copy(node->ctx, node->want);
	if (e1)
		new_want = isl_basic_map_apply_domain(node->ctx, new_want,
			    isl_basic_map_copy(node->ctx, e1->relation));
	if (e2)
		new_want = isl_basic_map_apply_range(node->ctx, new_want,
			    isl_basic_map_copy(node->ctx, e2->relation));

	eq_node *child = new eq_node(node->ctx, comp1, comp2, new_want);
	return child;
}

static isl_map *got_from_propagation(eq_node *node, edge *e1, edge *e2)
{
	isl_map *new_got = isl_map_copy(node->ctx, node->got);
	if (e1)
		new_got = isl_map_apply_domain(node->ctx, new_got,
				isl_map_from_basic_map(node->ctx,
				isl_basic_map_reverse(node->ctx,
				isl_basic_map_copy(node->ctx, e1->relation))));
	if (e2)
		new_got = isl_map_apply_range(node->ctx, new_got,
				isl_map_from_basic_map(node->ctx,
				isl_basic_map_reverse(node->ctx,
				isl_basic_map_copy(node->ctx, e2->relation))));
	return new_got;
}

void equivalence_checker::handle_copy(eq_node *node, int i)
{
	std::vector<edge *>::iterator iter;
	computation *copy = node->comp[i];
	isl_basic_map *want = node->want;
	isl_map *got;

	got = isl_map_empty(node->ctx, want->nparam, want->n_in, want->n_out);

	for (iter = copy->edges.begin(); iter != copy->edges.end(); ++iter) {
		edge *e = *iter;
		eq_node *child;
		if (i == 0)
			child = propagation_node(node, e, NULL);
		else
			child = propagation_node(node, NULL, e);

		if (!child)
			continue;

		handle(child);

		isl_map *new_got;
		if (i == 0)
			new_got = got_from_propagation(child, e, NULL);
		else
			new_got = got_from_propagation(child, NULL, e);
		dismiss(child);

		got = isl_map_union_disjoint(node->ctx, got, new_got);
	}
	node->got = got;
}

/* Compute the relation for which equivalence of the node holds.
 *
 * First, for function of zero arity (i.e., constants), equivalence
 * always holds.
 *
 * For functions of positive arity, we pair off each outgoing edge
 * of argument position j of the first computation (stored in *iter[0][j])
 * to either each edge of argument position j (for non-commutative operations)
 * or each edge (for commutative operations) of the second computation
 * (stored in *iter[1][j]).
 * We perform a backtracking search through all possible sets of matchings.
 * In each even level, we select an edge for the first computation.
 * In each odd level, we select an edge for the second computation.
 * For any set of matchings, we compute the intersection of the obtained
 * relations (got).i  If a partial set of matchings alread results in an
 * empty relation, we don't consider any of its extensions.
 * The final obtained relation is the union of the obtained relations
 * of all sets of matchings.
 */
void equivalence_checker::handle_propagation_same(eq_node *node)
{
	unsigned r = node->comp[0]->arity;
	if (r == 0) {
		node->got = isl_map_from_basic_map(node->ctx,
				isl_basic_map_copy(node->ctx, node->want));
		return;
	}
	std::vector<edge *>::iterator iter[2][r];
	std::vector<edge *>::iterator begin[2];
	std::vector<edge *>::iterator end[2];
	isl_basic_map *want = node->want;
	isl_map *got_at[r];
	isl_map *got;
	int level;
	bool is_comm = is_commutative(node->comp[0]->operation);
	std::map<pair<edge *, edge *>, isl_map *> cache;
	std::map<pair<edge *, edge *>, eq_node *> cache_child;

	for (int i = 0; i < 2; ++i) {
		begin[i] = node->comp[i]->edges.begin();
		end[i] = node->comp[i]->edges.end();
		for (int j = 0; j < r; ++j)
			iter[i][j] = begin[i];
	}
	for (int i = 0; i < r; ++i)
		got_at[i] = NULL;
	got = isl_map_empty(node->ctx, want->nparam, want->n_in, want->n_out);

	level = 0;
	while (level >= 0) {
		if (iter[level % 2][level / 2] == end[level % 2]) {
			iter[level % 2][level / 2] = begin[level % 2];
			--level;
			if (level >= 0)
				++iter[level % 2][level / 2];
			continue;
		}
		if (level % 2 == 0 && (*iter[0][level / 2])->pos != level / 2) {
			++iter[0][level / 2];
			continue;
		}
		if (level % 2 == 1) {
			if (!is_comm) {
				/* argument positions need to match */
				if ((*iter[1][level / 2])->pos != level / 2) {
					++iter[1][level / 2];
					continue;
				}
			} else {
				int l;
				/* argument position needs to be different
				 * from already matched arguments.
				 */
				for (l = 0; l < level / 2; ++l)
					if ((*iter[1][l])->pos ==
					    (*iter[1][level/2])->pos)
						break;
				if (l != level / 2) {
					++iter[1][level / 2];
					continue;
				}
			}
		}
		if (level % 2 == 0) {
			++level;
			continue;
		}
		edge *e1 = *iter[0][level / 2];
		edge *e2 = *iter[1][level / 2];

		eq_node *child;
		isl_map *new_got = NULL;

		if (cache.find(pair<edge *, edge *>(e1,e2)) != cache.end()) {
			new_got = isl_map_copy(node->ctx,
					cache[pair<edge *, edge *>(e1,e2)]);
			child = cache_child[pair<edge *, edge *>(e1,e2)];
		} else {
			child = propagation_node(node, e1, e2);
			if (child) {
				handle(child);
				new_got = got_from_propagation(child, e1, e2);
			}
			cache[pair<edge *, edge *>(e1,e2)] =
					    isl_map_copy(node->ctx, new_got);
			cache_child[pair<edge *, edge *>(e1,e2)] = child;
		}

		if (!child) {
			++iter[level % 2][level / 2];
			continue;
		}

		isl_map_free(node->ctx, got_at[level / 2]);
		got_at[level / 2] = new_got;
		if (level / 2 != 0)
			got_at[level / 2] = isl_map_intersect(node->ctx,
				got_at[level / 2],
				isl_map_copy(node->ctx, got_at[level / 2 - 1]));

		if (isl_map_is_empty(node->ctx, got_at[level/2])) {
			++iter[level % 2][level / 2];
			continue;
		}

		if (level / 2 != r - 1)
			++level;
		else {
			isl_map *copy = isl_map_copy(node->ctx, got_at[r - 1]);
			if (is_comm)
				got = isl_map_union(node->ctx, got, copy);
			else
				got = isl_map_union_disjoint(node->ctx, got,
									copy);
			++iter[level % 2][level / 2];
		}

		std::set<eq_node *>::iterator i;
		for (i = child->assumed.begin(); i != child->assumed.end(); ++i)
			node->assumed.insert(*i);
	}
	for (int i = 0; i < r; ++i)
		isl_map_free(node->ctx, got_at[i]);
	node->got = got;

	std::map<pair<edge *, edge *>, isl_map *>::iterator i;
	for (i = cache.begin(); i != cache.end(); ++i)
		isl_map_free(node->ctx, (*i).second);
	std::map<pair<edge *, edge *>, eq_node *>::iterator j;
	for (j = cache_child.begin(); j != cache_child.end(); ++j)
		dismiss((*j).second);
}

/* Perform induction, if possible, and widening if we have to.
 */
void equivalence_checker::handle_with_ancestor(eq_node *node, eq_node *a)
{
	if (a->narrowing ||
	    isl_basic_map_is_subset(node->ctx, node->want, a->want)) {
		node->got = isl_map_from_basic_map(node->ctx,
			isl_basic_map_intersect(node->ctx,
			    isl_basic_map_copy(node->ctx, node->want),
			    isl_basic_map_copy(node->ctx, a->want)));
		node->assumed.insert(a);
		a->need = isl_map_union(node->ctx, a->need,
				isl_map_copy(node->ctx, node->got));
	} else {
		isl_map *wants;
		isl_basic_map *aff;
		wants = isl_basic_map_union(node->ctx,
			    isl_basic_map_copy(node->ctx, node->want),
			    isl_basic_map_copy(node->ctx, a->want));
		aff = isl_map_affine_hull(node->ctx, wants);

		eq_node *child = new eq_node(node->ctx, node->comp[0],
						node->comp[1], aff);
		child->widening = 1;
		handle(child);
		node->got = isl_map_intersect(node->ctx,
				isl_map_copy(node->ctx, child->got),
				isl_map_from_basic_map(node->ctx,
				isl_basic_map_copy(node->ctx, node->want)));
		std::set<eq_node *>::iterator i;
		for (i = child->assumed.begin(); i != child->assumed.end(); ++i)
			node->assumed.insert(*i);
		dismiss(child);
	}
}

/* Construct and handle narrowing nodes for the given node.
 *
 * If the node itself was already a narrowing node, then we
 * simply return the empty relation.
 * Otherwise, we consider each basic relation in the obtained
 * relation, construct a new node with that basic relation as
 * requested relation and take the union of all obtained relations.
 */
void equivalence_checker::handle_narrowing(eq_node *node)
{
	node->reset = 1;
	node->assumed.clear();

	isl_basic_map *want = node->want;
	isl_map *new_got = isl_map_empty(node->ctx,
				want->nparam, want->n_in, want->n_out);
	if (node->narrowing) {
		isl_map_free(node->ctx, node->got);
		node->got = new_got;
		return;
	}

	for (int i = 0; i < node->got->n; ++i) {
		eq_node *child;
		child = new eq_node(node->ctx, node->comp[0],
			    node->comp[1],
			    isl_basic_map_copy(node->ctx, node->got->p[i]));
		child->narrowing = 1;
		handle(child);
		new_got = isl_map_union(node->ctx, new_got,
				isl_map_copy(node->ctx, child->got));

		std::set<eq_node *>::iterator i;
		for (i = child->assumed.begin(); i != child->assumed.end(); ++i)
			node->assumed.insert(*i);
		dismiss(child);
	}

	isl_map_free(node->ctx, node->got);
	node->got = new_got;
}

static int check_equivalence_array(isl_ctx *ctx, equivalence_checker *ec,
	dependence_graph *dg1, dependence_graph *dg2, int array)
{
	const char *array_name = dg1->output_arrays[array];
	isl_set *out1 = isl_set_copy(ctx, dg2->out->domain);
	isl_set *out2 = isl_set_copy(ctx, dg2->out->domain);
	out1 = isl_set_fix_dim_si(ctx, out1, out1->dim-1, array);
	out2 = isl_set_fix_dim_si(ctx, out2, out2->dim-1, array);
	int equal = isl_set_is_equal(ctx, out1, out2);
	isl_set_free(ctx, out1);
	assert(equal >= 0);
	if (!equal) {
		fprintf(stderr, "different output domains for array %s\n",
				array_name);
		return -1;
	}

	isl_map *id = isl_map_identity(ctx, out2->nparam, out2->dim);
	id = isl_map_intersect_range(ctx, id, out2);

	isl_set *left = isl_map_range(ctx, isl_map_copy(ctx, id));
	for (int i = 0; i < id->n; ++i) {
		if (isl_basic_map_is_empty(ctx, id->p[i]))
			continue;
		eq_node *root = new eq_node(ctx, dg1->out, dg2->out,
					isl_basic_map_copy(ctx, id->p[i]));
		ec->handle(root);
		if (!isl_map_is_empty(ctx, root->got)) {
			isl_set *range = isl_map_range(ctx,
						isl_map_copy(ctx, root->got));
			fprintf(stdout, "Equivalence of %s proved for\n",
								array_name);
			isl_set_dump(ctx, range, stdout, 0);
			left = isl_set_subtract(ctx, left, range);
		}
		ec->dismiss(root);
	}
	isl_map_free(ctx, id);
	if (!isl_set_is_empty(ctx, left)) {
		fprintf(stdout, "Equivalence of %s NOT proved for\n",
								array_name);
		isl_set_dump(ctx, left, stdout, 0);
	}
	isl_set_free(ctx, left);
}

static int check_equivalence(isl_ctx *ctx,
	dependence_graph *dg1, dependence_graph *dg2, arguments *arg)
{
	if (dg1->output_arrays.size() != dg2->output_arrays.size()) {
		fprintf(stderr, "different number of output arrays\n");
		return -1;
	}
	for (int i = 0; i < dg1->output_arrays.size(); ++i)
		if (strcmp(dg1->output_arrays[i], dg2->output_arrays[i])) {
			fprintf(stderr, "different output array %d: %s vs %s\n",
				dg1->output_arrays[i],
				dg2->output_arrays[i]);
			return -1;
		}

	equivalence_checker ec(arg);
	for (int i = 0; i < dg1->output_arrays.size(); ++i)
		check_equivalence_array(ctx, &ec, dg1, dg2, i);

	return 0;
}

int main(int argc, char *argv[])
{
	struct arguments arg;
	static struct argp argp = { argp_options, parse_opt, args_doc, 0, NULL };
	struct isl_ctx *ctx;
	dependence_graph *dg[2];

	set_program_name(argv[0]);
	argp_parse(&argp, argc, argv, 0, 0, &arg);

	ctx = isl_ctx_alloc();
	if (!ctx) {
		fprintf(stderr, "Unable to allocate ctx\n");
		return -1;
	}

	for (int i = 0; i < 2; ++i) {
		FILE *in;
		pdg::PDG *pdg;
		in = fopen(arg.program[i], "r");
		if (!in) {
			fprintf(stderr, "Unable to open %s\n", arg.program[i]);
			return -1;
		}
		pdg = yaml::Load<pdg::PDG>(in);
		fclose(in);
		if (!pdg) {
			fprintf(stderr, "Unable to read %s\n", arg.program[i]);
			return -1;
		}

		dg[i] = pdg_to_dg(ctx, pdg);
		pdg->free();
		delete pdg;
	}

	int res = check_equivalence(ctx, dg[0], dg[1], &arg);

	for (int i = 0; i < 2; ++i)
		delete dg[i];
	isl_ctx_free(ctx);

	return res;
}
