#include <barvinok/barvinok.h>
#include "argp.h"
#undef value_compare
#include "yaml.h"
#include "pdg.h"
#include "da.h"
#include "evalue_util.h"
#include "clast_scanner.h"
extern "C" {
#include "piputil.h"
}
#include <cloog/cloog.h>
#include "cloog_util.h"

#define ALLOC(type) (type*)malloc(sizeof(type))

using pdg::PDG;
using namespace da;

#define OPT_WITH_INPUT	    (BV_OPT_LAST+1)

struct argp_option argp_options[] = {
    { "with-input",	OPT_WITH_INPUT,		0,	0,
	"include size of input arrays in total" },
    { 0 }
};

struct options {
    barvinok_options	*barvinok;
    int			 with_input;
};

static error_t parse_opt(int key, char *arg, struct argp_state *state)
{
    struct options *options = (struct options *)(state->input);

    switch (key) {
    case ARGP_KEY_INIT:
	state->child_inputs[0] = options->barvinok;
	options->with_input = 0;
	break;
    case OPT_WITH_INPUT:
	options->with_input = 1;
	break;
    default:
	return ARGP_ERR_UNKNOWN;
    }
    return 0;
}

static void dump_program(PDG *pdg, int *output, struct options *options);
static void scan_program(PDG *pdg, evalue *inputsize,
			 CloogProgram * p, CloogOptions *options);

int main(int argc, char * argv[])
{
    PDG *pdg;
    static struct argp_child argp_children[] = {
	{ &barvinok_argp,    	0,	"barvinok options",	1 },
	{ 0 }
    };
    static struct argp argp = { argp_options, parse_opt, 0, 0, argp_children };
    struct barvinok_options *bv_options = barvinok_options_new_with_defaults();
    struct options options;

    options.barvinok = bv_options;
    argp_parse(&argp, argc, argv, 0, 0, &options);

    pdg = yaml::Load<PDG>(stdin);
    int output[pdg->arrays.size()];

    for (int i = 0; i < pdg->arrays.size(); ++i) {
	int n = pdg->dependences.size();
	find_deps(pdg, pdg->arrays[i], data_reuse);
	output[i] = n == pdg->dependences.size();
    }

    dump_program(pdg, output, &options);
    pdg->free();
    delete pdg;
    barvinok_options_free(bv_options);

    return 0;
}

static char ** names_generate_items(int nb_items, const char * prefix)
{ int i ;
  char ** names ;
    
  names = (char **)malloc(nb_items*sizeof(char *)) ;
  if (names == NULL) 
  { fprintf(stderr, "[CLooG]ERROR: memory overflow.\n") ;
    exit(1) ;
  }
  for (i=0;i<nb_items;i++)
  { names[i] = (char *)malloc(MAX_NAME*sizeof(char)) ;
    if (names[i] == NULL) 
    { fprintf(stderr, "[CLooG]ERROR: memory overflow.\n") ;
      exit(1) ;
    }
    if (prefix == NULL)
    sprintf(names[i],"%d",i) ;
    else
    sprintf(names[i],"%s%d",prefix,i) ;
  }
  
  return names ;
}

static CloogLoop *poly2cloog(Polyhedron *P, pdg::node *node, pdg::access *ac,
				int nr, unsigned MaxRays)
{
    Matrix *M;
    Polyhedron *Q;
    CloogStatement * statement;
    CloogLoop *l = cloog_loop_malloc();
    int extra = 0;
    int fulldim = node->prefix.size() + 1;
    l->next = NULL;
    l->inner = NULL;
    for (int i = 0; i < node->prefix.size(); ++i)
	if (node->prefix[i] != -1)
	    ++extra;
    ++extra;
    M = Matrix_Alloc(P->NbConstraints + extra, P->Dimension + 2 + extra);
    for (int i = 0, j = 0; i < node->prefix.size(); ++i) {
	if (node->prefix[i] == -1)
	    continue;
	value_set_si(M->p[j][1+i], -1);
	value_set_si(M->p[j][P->Dimension+1+extra], node->prefix[i]);
	++j;
    }
    value_set_si(M->p[extra-1][fulldim], -1);
    value_set_si(M->p[extra-1][P->Dimension+1+extra], ac->nr);
    for (int i = 0; i < P->NbConstraints; ++i) {
	value_assign(M->p[extra+i][0], P->Constraint[i][0]);
	for (int j = 0, k = 0; j < node->prefix.size(); ++j) {
	    if (node->prefix[j] != -1)
		continue;
	    value_assign(M->p[extra+i][1+j], P->Constraint[i][1+k]);
	    ++k;
	}
	Vector_Copy(P->Constraint[i]+1+fulldim-extra, M->p[extra+i]+1+fulldim,
		    P->Dimension+1-(fulldim-extra));
    }
    Q = Constraints2Polyhedron(M, MaxRays);
    Matrix_Free(M);
    l->domain = cloog_domain_alloc(Q);

    statement = cloog_statement_alloc(nr);
    l->block = cloog_block_alloc(statement, NULL, 0, NULL, Q->Dimension);
    return l;
}

/* Compute the number of array elements that are read but not
 * (previously) written.  These are assumed to be the input.
 */
static evalue *compute_input_size(PDG* pdg, struct options *options)
{
    evalue *inputsize = evalue_zero();

    if (!options->with_input)
	return inputsize;

    unsigned nparam = pdg->params.size();

    for (int j = 0; j < pdg->dependences.size(); ++j) {
	pdg::dependence *dep = pdg->dependences[j];
	if (dep->type != pdg::dependence::uninitialized)
	    continue;
	pdg::UnionMap *rel = dep->relation;
	Polyhedron *D = *rel;
	for (Polyhedron *P = D; P; P = P->next) {
	    evalue *EP = barvinok_enumerate_e_with_options(P, 
			    P->Dimension - rel->output - nparam, nparam,
			    options->barvinok);
	    eadd(EP, inputsize);
	    evalue_free(EP);
	}
	Domain_Free(D);
    }

    return inputsize;
}

static void dump_program(PDG *pdg, int *output, struct options *options)
{
    int nparam = pdg->params.size();
    Polyhedron *ctx = pdg->context ? (Polyhedron *) *pdg->context : 
				     Universe_Polyhedron(nparam);
    CloogOptions *cloog_options = cloog_options_malloc();
    cloog_options->compilable = 1;
    CloogProgram * p = cloog_program_malloc();
    p->context = cloog_domain_alloc(ctx);
    int maxdim = 0;
    evalue *inputsize;

    CloogNames * names = cloog_names_malloc();

    p->names = names;
    p->nb_scattdims = 0;
    p->scaldims = NULL;

    inputsize = compute_input_size(pdg, options);
    evalue_simplify_domains(inputsize, ctx);
    //print_evalue(stderr, inputsize, NULL);

    CloogLoop **next = &p->loop;
    CloogBlockList **nextBL = &p->blocklist;
    for (int i = 0; i < pdg->nodes.size(); ++i) {
	pdg::node *node = pdg->nodes[i];
	pdg::statement *s = pdg->nodes[i]->statement;
	for (int j = 0; j < pdg->dependences.size(); ++j) {
	    pdg::dependence *dep = pdg->dependences[j];
	    if (dep->to != pdg->nodes[i] && dep->from != pdg->nodes[i])
		continue;
	    if (!options->with_input &&
		    dep->type == pdg::dependence::uninitialized)
		continue;
	    pdg::UnionMap *rel = dep->relation;
	    Polyhedron *D = *rel;
	    for (Polyhedron *P = D; P; P = P->next) {
		if (dep->to == pdg->nodes[i]) {
		    Polyhedron *read = Polyhedron_projectout_dep(P, rel->input);
		    *next = poly2cloog(read, dep->to, dep->to_access, -1,
					options->barvinok->MaxRays);
		    Polyhedron_Free(read);
		    *nextBL = cloog_block_list_alloc((*next)->block);

		    if ((*next)->domain->polyhedron->Dimension > maxdim)
			maxdim = (*next)->domain->polyhedron->Dimension;

		    next = &(*next)->next;
		    nextBL = &(*nextBL)->next;
		}
		if (dep->from == pdg->nodes[i]) {
		    *next = poly2cloog(P, dep->from, dep->from_access, 1,
					options->barvinok->MaxRays);
		    *nextBL = cloog_block_list_alloc((*next)->block);

		    if ((*next)->domain->polyhedron->Dimension > maxdim)
			maxdim = (*next)->domain->polyhedron->Dimension;

		    next = &(*next)->next;
		    nextBL = &(*nextBL)->next;
		}
	    }
	    Domain_Free(D);
	}
	for (int j = 0; j < s->accesses.size(); ++j) {
	    pdg::access *access = s->accesses[j];
	    if (access->type == pdg::access::read)
		continue;
	    int k;
	    for (k = 0; k < pdg->arrays.size(); ++k)
		if (pdg->arrays[k] == access->array)
		    break;
	    if (!output[k])
		continue;
	    Polyhedron *D = *node->source;
	    for (Polyhedron *P = D; P; P = P->next) {
		*next = poly2cloog(P, node, access, 1,
					options->barvinok->MaxRays);
		*nextBL = cloog_block_list_alloc((*next)->block);

		if ((*next)->domain->polyhedron->Dimension > maxdim)
		    maxdim = (*next)->domain->polyhedron->Dimension;

		next = &(*next)->next;
		nextBL = &(*nextBL)->next;
	    }
	    Domain_Free(D);
	}
    }

    names->nb_parameters = nparam;
    names->parameters = (char**)malloc(nparam * sizeof(char*));
    for (int i = 0; i < nparam; ++i)
	names->parameters[i] = strdup(pdg->params[i]->name->s.c_str());
    names->nb_iterators = maxdim;
    names->iterators = names_generate_items(maxdim, "i");
    names->scattering = NULL;
    names->nb_scattering = 0;
    names->nb_scalars = 0;

    p = cloog_program_generate(p, cloog_options);
    scan_program(pdg, inputsize, p, cloog_options);
    cloog_options_free(cloog_options);
    cloog_program_free(p);
    free_evalue_refs(inputsize);
    free(inputsize);
}

struct mem_counter : public clast_scanner {
    char tmpfile[L_tmpnam];
    char buffer[L_tmpnam+40];

    int count();
    void print_user(clast_user_stmt *s);
    void print_full(PDG *pdg, evalue *inputsize, CloogNames* names, clast_stmt *s);
};

void mem_counter::print_full(PDG *pdg, evalue *inputsize,
			     CloogNames* names, clast_stmt *s)
{
    dst = stdout;
    print_header();
    fprintf(dst, "void main() {\n");
    indent += 4;
    print_indent();
    if (names->nb_iterators <= 0)
	return;
    fprintf(dst, "int ");
    for (int i = 0; i < names->nb_iterators; ++i) {
	if (i)
	    fprintf(dst, ", ");
	fprintf(dst, names->iterators[i]);
    }
    fprintf(dst, ";\n");
    print_indent();
    fprintf(dst, "long count = %s;\n", 
		    evalue2string(inputsize, pdg->params).c_str());
    print_indent();
    fprintf(dst, "long max = count;\n");

    print(s);

    print_indent();
    fprintf(dst, "printf(\"final: %%d\\n\", count);\n");
    fprintf(dst, "printf(\"max: %%d\\n\", max);\n");
    fprintf(dst, "}\n");
}

void mem_counter::print_user(clast_user_stmt *u)
{
    print_indent();
    if (u->statement->number == 1) {
	fprintf(dst, "if (++count > max)\n");
	print_indent();
	fprintf(dst, "    max = count;\n");
    } else
	fprintf(dst, "--count;\n");
}

static void scan_program(PDG *pdg, evalue *inputsize,
			 CloogProgram * p, CloogOptions *options)
{
    clast_stmt *stmt = cloog_util_create_clast(p, options);
    mem_counter mc;
    mc.print_full(pdg, inputsize, p->names, stmt);
    cloog_clast_free(stmt);
}
