#include <polylib/polylibgmp.h>
#include <stdio.h>  /* needed for piplib ! */
#include <piplib/piplibMP.h>
#include <assert.h>
#include "combine.h"
#include "piputil.h"
#include "isl_set.h"
#include "isl_map.h"
#include "isl_set_polylib.h"
#include "isl_map_polylib.h"

#ifndef POL_INTEGER
#define POL_INTEGER	0
#endif

#define POL_HAS(P,f)            (F_ISSET(P, f))
#define POL_SET(P,f)            (F_SET(P, f))

#define MAXRAYS    (POL_NO_DUAL | POL_INTEGER)

static PipMatrix *poly2pip(Polyhedron *P, int pos, int n, int nparam)
{
    PipMatrix *	    matrix;
    int i, j;
    unsigned int extra = P->Dimension - pos - n - nparam;

    matrix = pip_matrix_alloc(P->NbConstraints, P->Dimension+2);
    for (i = 0; i < P->NbConstraints; ++i) {
	value_assign(matrix->p[i][0], P->Constraint[i][0]);
	for (j = 0; j < n; ++j)
	    value_assign(matrix->p[i][1+j], P->Constraint[i][1+pos+j]);
	for (j = 0; j < pos; ++j)
	    value_assign(matrix->p[i][1+n+j], P->Constraint[i][1+j]);
	for (j = 0; j < extra; ++j)
	    value_assign(matrix->p[i][1+n+pos+j], P->Constraint[i][1+n+pos+j]);
	for (j = 1+pos+n+extra; j < P->Dimension+2; ++j)
	    value_assign(matrix->p[i][j], P->Constraint[i][j]);
    }

    return matrix;
}

/* remove existential variables that are affine combinations of
 * other existential variables
 *
 * pos: position of first existential variable
 * n: number of existential variables
 */
static Polyhedron *remove_redundant_existentials(Polyhedron *P, int pos, int n)
{
    int i, r, s;
    Matrix M, *M2, *M3;
    Polyhedron *R;
    for (i = 0; i < n; ++i) {
	for (r = 0; r < P->NbEq; ++r)
	    if (First_Non_Zero(P->Constraint[r]+1, pos-1) == -1 &&
		value_notzero_p(P->Constraint[r][pos+i]))
		break;
	if (r == P->NbEq)
	    continue;
	if (value_notone_p(P->Constraint[r][pos+i]) &&
	    value_notmone_p(P->Constraint[r][pos+i]))
	    continue;
	if (r == P->NbEq)
	    continue;
	for (s = 0; s < P->NbConstraints; ++s) {
	    if (s == r)
		continue;
	    if (value_zero_p(P->Constraint[s][pos+i]))
		continue;
	    Combine(P->Constraint[s], P->Constraint[r],
		    P->Constraint[s], pos+i, P->Dimension+1);
	}
	/* Matrix "view" of P for RemoveColumn */
	M.NbRows = P->NbConstraints;
	M.NbColumns = P->Dimension+2;
	M.p_Init = P->p_Init;
	M.p = P->Constraint;
	M2 = RemoveColumn(&M, pos+i);
	M3 = RemoveRow(M2, r);
	Matrix_Free(M2);
	R = Constraints2Polyhedron(M3, MAXRAYS);
	Matrix_Free(M3);
	Polyhedron_Free(P);
	P = R;
	/* recheck all existential variables */
	i = -1;
	--n;
    }
    return P;
}

static Polyhedron *D_remove_redundant_existentials(Polyhedron *D, int dim, 
						   int nparam)
{
    Polyhedron *P, *N, *R = NULL;
    for (P = D; P; P = N) {
	N = P->next;
	P->next = NULL;
	P = remove_redundant_existentials(P, 1+dim, P->Dimension-dim-nparam);
	P->next = R;
	R = P;
    }
    return R;
}

static void Matrix_Shift_First_Columns(Value **M, int rows, int n, int pos)
{
    int i;
    Vector *shifted = Vector_Alloc(n);
    for (i = 0; i < rows; ++i) {
	Vector_Copy(M[i]+1, shifted->p, n);
	Vector_Copy(M[i]+1+n, M[i]+1, pos);
	Vector_Copy(shifted->p, M[i]+1+pos, n);
    }
    Vector_Free(shifted);
}

/* Move first n columns to position pos, moving all columns up to pos+n-1 down
 * by n
 */
static void Polyhedron_Shift_First_Columns(Polyhedron *P, int n, int pos)
{
    int i;
    if (pos == 0)
	return;
    if (POL_HAS(P, POL_INEQUALITIES)) {
	Matrix M;
	Matrix_Shift_First_Columns(P->Constraint, P->NbConstraints, n, pos);
	/* Matrix "view" of P for Gauss */
	M.NbRows = P->NbConstraints;
	M.NbColumns = P->Dimension+2;
	M.p_Init = P->p_Init;
	M.p = P->Constraint;
	/* Put constraints back in "standard" form */
	Gauss(&M, P->NbEq, P->Dimension+1);
    }
    if (POL_HAS(P, POL_POINTS))
	Matrix_Shift_First_Columns(P->Ray, P->NbRays, n, pos);
    if (P->next)
	Polyhedron_Shift_First_Columns(P->next, n, pos);
}

Polyhedron *pip_lexminmax(Polyhedron *P, int pos, int n, int nparam, int max)
{
    unsigned int nvar = P->Dimension - nparam;
    Polyhedron  *min;
    struct isl_ctx *ctx;
    struct isl_basic_map *bmap;
    struct isl_basic_set *dom;
    struct isl_map *min_map;

    POL_ENSURE_INEQUALITIES(P);
    P = Polyhedron_Copy(P);
    Polyhedron_Shift_First_Columns(P, pos, n);

    ctx = isl_ctx_alloc();
    bmap = isl_basic_map_new_from_polylib(ctx, P, nparam, n, nvar - n);
    Polyhedron_Free(P);

    bmap = isl_basic_map_reverse(ctx, bmap);
    dom = isl_basic_set_alloc(ctx, nparam, nvar - n, 0, 0, 0);
    if (max)
	min_map = isl_basic_map_lexmax(ctx, bmap, dom, NULL);
    else
	min_map = isl_basic_map_lexmin(ctx, bmap, dom, NULL);
    min_map = isl_map_reverse(ctx, min_map);
    assert(min_map);
    min = isl_map_to_polylib(ctx, min_map);
    isl_map_free(ctx, min_map);

    isl_ctx_free(ctx);

    Polyhedron_Shift_First_Columns(min, n, pos);

    return min;
}

Polyhedron *Polyhedron_projectout(Polyhedron *P, int pos, int n, int nparam)
{
    unsigned int nvar = P->Dimension - nparam;
    Polyhedron  *min;
    struct isl_ctx *ctx;
    struct isl_basic_map *bmap;
    struct isl_basic_set *bset;
    struct isl_set *min_set;

    POL_ENSURE_INEQUALITIES(P);
    P = Polyhedron_Copy(P);
    Polyhedron_Shift_First_Columns(P, pos, n);

    ctx = isl_ctx_alloc();
    bmap = isl_basic_map_new_from_polylib(ctx, P, nparam, n, nvar - n);
    Polyhedron_Free(P);

    bset = isl_basic_map_range(ctx, bmap);
    min_set = isl_basic_set_compute_divs(ctx, bset);
    min = isl_set_to_polylib(ctx, min_set);
    isl_set_free(ctx, min_set);

    isl_ctx_free(ctx);

    return min;
}

Polyhedron *Domain_projectout(Polyhedron *D, int pos, int n, int nparam)
{
    Polyhedron *P, *R = NULL;
    for (P = D; P; P = P->next)
	R = Domain_concat(Polyhedron_projectout(P, pos, n, nparam), R);
}

static int quast_empty(PipQuast *q)
{
    return !q->newparm && !q->condition && !q->list;
}

static int return_dim(void *first, void *second)
{
    return 2*(int)first - 1;
}

/* The variables in the mapping are 
 *	previous_user next_user extra parameters
 * previous_user vars are constrained using equalities in M
 *
 * The variables in the remaining domains are
 *	first_user extra parameters
 */
struct pip_dep_result pip_inputdep(Polyhedron *D, int dim, Matrix *M)
{
    int nparam = D->Dimension - dim;
    struct pip_dep_result res;
    struct pip_dependence_info *dep;
    struct pip_rel *pr;
    Polyhedron *rel;
    Matrix *M2;
    int i;

    M2 = Matrix_Alloc(M->NbRows + D->NbConstraints, 1+dim+M->NbRows+nparam+1);
    for (i = 0; i < M->NbRows; ++i) {
	Vector_Copy(M->p[i], M2->p[i]+1, M->NbColumns-nparam-1);
	Vector_Copy(M->p[i]+M->NbColumns-nparam-1, 
		    M2->p[i]+1+dim+M->NbRows, nparam+1);
	value_set_si(M2->p[i][1+dim+i], -1);
    }
    for (i = 0; i < D->NbConstraints; ++i) {
	Vector_Copy(D->Constraint[i], M2->p[M->NbRows+i], 1+dim);
	Vector_Copy(D->Constraint[i]+1+dim, 
		    M2->p[M->NbRows+i]+1+dim+M->NbRows, nparam+1);
    }
    rel = Constraints2Polyhedron(M2, MAXRAYS);
    Matrix_Free(M2);
    dep = (struct pip_dependence_info *) 
	    malloc(sizeof(struct pip_dependence_info) + sizeof(struct pip_rel));
    dep->ctx = Domain_projectout(D, 0, dim, nparam);
    assert(dep->ctx->Dimension == nparam);
    dep->n_write = 1;
    dep->precedes_level = return_dim;
    dep->read.n_input = dim;
    dep->read.n_output = M->NbRows;
    dep->read.data = (void *)dim;
    dep->read.rel = rel;
    dep->write[0].n_input = dim;
    dep->write[0].n_output = M->NbRows;
    dep->write[0].data = (void *)dim;
    dep->write[0].rel = rel;

    pr = pip_lastwrite(dep);

    res.D = pr[1].rel;
    res.M = pr[0].rel;

    free(pr);
    Domain_Free(dep->ctx);
    Domain_Free(rel);
    free(dep);

    return res;
}

static poly2pipc(Polyhedron *P, PipMatrix *matrix,
		 int prow, int nrow, int mrow,
		 int pcol, int ncol, int mcol)
{
    int i, j;
    for (i = 0; i < nrow; ++i)
	for (j = 0; j < ncol; ++j)
	    value_assign(matrix->p[mrow+i][mcol+j], P->Constraint[prow+i][pcol+j]);
}

static sub_poly2pipc(Polyhedron *P, PipMatrix *matrix,
		     int prow, int nrow, int mrow,
		     int pcol, int ncol, int mcol)
{
    int i, j;
    for (i = 0; i < nrow; ++i)
	for (j = 0; j < ncol; ++j)
	    value_subtract(matrix->p[mrow+i][mcol+j], matrix->p[mrow+i][mcol+j], 
			   P->Constraint[prow+i][pcol+j]);
}

/* add constraints to matrix starting from row specifying that
 * the n coordinates starting at pos1 are equal to those starting at pos2
 */
static void var_equal(PipMatrix *matrix, int pos1, int n, int pos2, int row)
{
    int i;
    for (i = 0; i < n; ++i) {
	value_set_si(matrix->p[row+i][0], 0);
	value_set_si(matrix->p[row+i][pos1+i], -1);
	value_set_si(matrix->p[row+i][pos2+i], 1);
	value_set_si(matrix->p[row+i][matrix->NbColumns-1], 0);
    }
}

/* add constraints to matrix starting from row specifying that
 * the coordinate at pos1 is smaller than the one at pos2
 */
static void var_smaller(PipMatrix *matrix, int pos1, int pos2, int row)
{
    value_set_si(matrix->p[row][0], 1);
    value_set_si(matrix->p[row][pos1], -1);
    value_set_si(matrix->p[row][pos2], 1);
    value_set_si(matrix->p[row][matrix->NbColumns-1], -1);
}

static struct isl_basic_map *before_at_level(struct isl_ctx *ctx,
		unsigned nparam, unsigned in, unsigned out, int level)
{
	if (level % 2)
		return isl_basic_map_less_at(ctx, nparam, in, out, level/2);
	else
		return isl_basic_map_equal(ctx, nparam, in, out, level/2);
}

/* Compute last iteration of write j that precedes a give read at level level */
static struct isl_map *last_write_2(struct isl_ctx *ctx,
				    struct pip_dependence_info *dep, 
				    struct isl_basic_set *set_C,
				    int j, int level, struct isl_set **empty)
{
    struct isl_map *map;
    struct isl_basic_map *read_map;
    struct isl_basic_map *write_map;
    struct isl_basic_map *dep_map;
    struct isl_basic_map *before;
    struct isl_map *result;
    int nparam = dep->ctx->Dimension;
    assert(dep->read.n_output == dep->write[j].n_output);
    assert(!dep->write[j].rel->next);

    read_map = isl_basic_map_new_from_polylib(ctx, dep->read.rel,
				nparam, dep->read.n_input, dep->read.n_output);
    write_map = isl_basic_map_new_from_polylib(ctx, dep->write[j].rel,
			nparam, dep->write[j].n_input, dep->write[j].n_output);
    write_map = isl_basic_map_reverse(ctx, write_map);
    dep_map = isl_basic_map_apply_range(ctx, read_map, write_map);
    before = before_at_level(ctx, nparam, dep->read.n_input,
				dep->write[j].n_input, level);
    dep_map = isl_basic_map_intersect(ctx, dep_map, before);
    result = isl_basic_map_lexmax(ctx, dep_map, set_C, empty);
    result = isl_map_reverse(ctx, result);

    return result;
}

/* PolyLib performs Gaussian elimination using the equalities
 * in a polyhedron.  The backsubstitution may result in some
 * of the variables appearing with coefficient different from
 * one, even though we know they should have a coefficient of
 * one.
 * This procedure tries to undo some of these backsubstitutions.
 */
static void ungauss(Polyhedron *map, int row)
{
    int i, j, k;
    int dim = map->Dimension;
    Value tmp, one;
    Value **p = map->Constraint;
    value_init(tmp);
    value_init(one);
    value_set_si(one, 1);
    for (j = row+1; j < dim; ++j) {
	value_modulus(tmp, p[row][1+j], p[row][1+row]);
	if (value_zero_p(tmp))
	    continue;
	for (i = row+1; i < map->NbEq; ++i) {
	    value_modulus(tmp, p[i][1+j], p[row][1+row]);
	    if (value_zero_p(tmp))
		continue;
	    k = First_Non_Zero(p[i]+1, dim);
	    value_modulus(tmp, p[i][1+k], p[row][1+row]);
	    if (value_notzero_p(tmp))
		continue;
	    value_modulus(tmp, p[i][1+j], p[row][1+j]);
	    if (value_notzero_p(tmp))
		continue;
	    value_division(tmp, p[i][1+j], p[row][1+j]);
	    value_oppose(tmp, tmp);
	    Vector_Combine(p[row]+1, p[i]+1, p[row]+1, one, tmp, dim+1);
	    Vector_Normalize(p[row]+1, dim+1);
	    break;
	}
	if (value_one_p(p[row][row+1]))
	    break;
    }
    assert(j < dim);
    value_clear(one);
    value_clear(tmp);
}

static void assert_dependent_mapping(Polyhedron *map, int dim)
{
    int i;
    assert(map->NbConstraints >= dim);
    for (i = 0; i < dim; ++i) {
	assert(value_zero_p(map->Constraint[i][0]));
	if (!value_one_p(map->Constraint[i][1+i]))
	    ungauss(map, i);
    }
}

/* Project out first dim variables that are assumed to be
 * affine combination of the other variables.
 */
Polyhedron *Polyhedron_projectout_dep(Polyhedron *map, int dim)
{
    Matrix *M;
    Polyhedron *P;
    int i;
    assert_dependent_mapping(map, dim);
    M = Matrix_Alloc(map->NbConstraints - dim, 2 + map->Dimension - dim);
    for (i = dim; i < map->NbConstraints; ++i) {
	value_assign(M->p[i-dim][0], map->Constraint[i][0]);
	assert(First_Non_Zero(map->Constraint[i]+1, dim) == -1);
	Vector_Copy(map->Constraint[i]+1+dim, M->p[i-dim]+1, 
		    map->Dimension - dim + 1);
    }
    P = Constraints2Polyhedron(M, MAXRAYS);
    Matrix_Free(M);
    return P;
}

/* Compute last iteration of write k that precedes a given read at level read_level,
 * but follows at level write_level the iteration of write j associated
 * to the read iteration by old.
 */
static struct isl_map *last_write_3(struct isl_ctx *ctx,
				    struct pip_dependence_info *dep,
				    struct isl_basic_map *old_map,
				    int j, int read_level,
				    int k, int write_level,
				    struct isl_set **empty)
{
    struct isl_basic_set *set_C;
    struct isl_basic_map *read_map;
    struct isl_basic_map *write_map;
    struct isl_basic_map *dep_map;
    struct isl_basic_map *after_write;
    struct isl_basic_map *before_read;
    struct isl_map *result;
    int nparam = dep->ctx->Dimension;
    assert(dep->read.n_output == dep->write[j].n_output);
    assert(dep->read.n_output == dep->write[k].n_output);
    assert(!dep->write[j].rel->next);
    assert(!dep->write[k].rel->next);

    set_C = isl_basic_map_range(ctx, isl_basic_map_copy(ctx, old_map));
    read_map = isl_basic_map_new_from_polylib(ctx, dep->read.rel,
				nparam, dep->read.n_input, dep->read.n_output);
    write_map = isl_basic_map_new_from_polylib(ctx, dep->write[k].rel,
			nparam, dep->write[k].n_input, dep->write[k].n_output);

    write_map = isl_basic_map_reverse(ctx, write_map);
    dep_map = isl_basic_map_apply_range(ctx, read_map, write_map);
    after_write = before_at_level(ctx, nparam, dep->write[k].n_input,
			dep->write[j].n_input, write_level);
    after_write = isl_basic_map_apply_range(ctx, after_write, old_map);
    after_write = isl_basic_map_reverse(ctx, after_write);
    dep_map = isl_basic_map_intersect(ctx, dep_map, after_write);
    before_read = before_at_level(ctx, nparam, dep->read.n_input,
				    dep->write[k].n_input, read_level);
    dep_map = isl_basic_map_intersect(ctx, dep_map, before_read);
    result = isl_basic_map_lexmax(ctx, dep_map, set_C, empty);
    result = isl_map_reverse(ctx, result);

    return result;
}

Polyhedron *EPoly_Intersection(Polyhedron *A, Polyhedron *B, int d, int nparam)
{
    int A_extra = A->Dimension - d - nparam;
    int B_extra = B->Dimension - d - nparam;
    Matrix *M = Matrix_Alloc(A->NbConstraints + B->NbConstraints, 
			     2 + A->Dimension + B_extra);
    Polyhedron *P;
    int i, A_row = A->NbConstraints;
    for (i = 0; i < A_row; ++i) {
	Vector_Copy(A->Constraint[i], M->p[i], 1+d+A_extra);
	Vector_Copy(A->Constraint[i]+1+d+A_extra, M->p[i]+1+d+A_extra+B_extra, 
		    nparam+1);
    }
    for (i = 0; i < B->NbConstraints; ++i) {
	Vector_Copy(B->Constraint[i], M->p[i+A_row], 1+d);
	Vector_Copy(B->Constraint[i]+1+d, M->p[i+A_row]+1+d+A_extra, 
		    B_extra+nparam+1);
    }
    P = Constraints2Polyhedron(M, MAXRAYS);
    Matrix_Free(M);
    /* return remove_redundant_existentials(P, 1+d, P->Dimension-d-nparam); */
    return P;
}

Polyhedron *Domain_concat(Polyhedron *head, Polyhedron *tail)
{
    Polyhedron *S;
    if (!head)
	return tail;
    for (S = head; S->next; S = S->next)
	;
    S->next = tail;
    return head;
}

/* Given a "read" access, a list of n "write" accesses and a context,
 * this function computes for each iteration of the read access,
 * the last write access of the list that wrote the data read by
 * the read access.
 * Each access is given by a relation between the loop iterators
 * and the array indices.
 * The result is a list of n+1 relations, the first n of which
 * correspond to the n write accesses and the last corresponds
 * to iterations of the read access that read data that is not
 * written by any of the n write accesses.
 *
 * The algorithm considers all levels in which a write may precede a read,
 * where a level may either be a statement level or a loop level.
 * The outermost statement level is 0, the first loop level is 1, etc...
 * The algorithm basically does the following:
 * for all levels l of the read access from innermost to outermost
 *	for all writes w that may precede the read access at that level
 *	    compute the last iteration of the write that precedes the read access
 *					    at that level
 *	    add result to possible last writes at level l of write w
 *	    for all writes w2 that we haven't considerd yet at this level that may
 *					    also precede the read access
 *		for all levels l2 of w from l to innermost
 *		    for all possible last writes dep of w at l
 *			compute last iteration of w2 between the write and read
 *								of dep
 *			add result to possible writes at level l of write w2
 *			and replace possible last write dep by remainder
 */
struct pip_rel *pip_lastwrite(struct pip_dependence_info *dep)
{
    struct isl_ctx *ctx;
    struct isl_basic_map *read_map;
    struct isl_basic_set *read_dom;
    struct isl_set *todo;
    int read_level, j, k, write_level, m;
    int readlen = 2 * dep->read.n_input + 1;
    struct pip_rel *final_rel;
    struct isl_map **temp_rel;
    Polyhedron *CA, *P;
    unsigned nparam = dep->ctx->Dimension;
    final_rel = (struct pip_rel *)malloc((dep->n_write+1)*sizeof(struct pip_rel));

    for (j = dep->n_write-1; j >=0; --j) {
	final_rel[j].n_input = dep->write[j].n_input;
	final_rel[j].n_output = dep->read.n_input;
	final_rel[j].data = dep->write[j].data;
	final_rel[j].rel = NULL;
    }
    final_rel[dep->n_write].n_input = 0;
    final_rel[dep->n_write].n_output = dep->read.n_input;
    final_rel[dep->n_write].data = NULL;
    final_rel[dep->n_write].rel = NULL;

    ctx = isl_ctx_alloc();
    assert(!dep->read.rel->next);
    CA = align_context(dep->ctx, dep->read.rel->Dimension, MAXRAYS);
    P = DomainIntersection(dep->read.rel, CA, MAXRAYS);
    Domain_Free(CA);
    read_map = isl_basic_map_new_from_polylib(ctx, P, dep->ctx->Dimension,
			dep->read.n_input, dep->read.n_output);
    Domain_Free(P);
    read_dom = isl_basic_map_domain(ctx, read_map);
    todo = isl_basic_set_compute_divs(ctx, read_dom);
    if (!todo)
	goto error;
    if (todo->n == 0)
	goto done;

    temp_rel = isl_alloc_array(ctx, struct isl_map *, dep->n_write);

    for (read_level = readlen-1; read_level >= 0; --read_level) {
	Polyhedron *rest = NULL, *P;

	for (j = dep->n_write-1; j >=0; --j)
	    temp_rel[j] = isl_map_empty(ctx, nparam,
				dep->write[j].n_input, dep->read.n_input);

	for (j = dep->n_write-1; j >=0; --j) {
	    struct isl_set *rest;
	    int writelen = 2 * dep->write[j].n_input + 1;
	    int plevel;
	    int i;

	    plevel = dep->precedes_level(dep->write[j].data, dep->read.data);
	    if (plevel < read_level)
		continue;
	    if (!(read_level % 2) && plevel > read_level)
		continue;

	    rest = isl_set_empty(ctx, nparam, dep->read.n_input);

	    for (i = 0; i < todo->n; ++i) {
		struct isl_map *T;
		struct isl_set *trest;
		struct isl_basic_set *copy =
			isl_basic_set_copy(ctx, todo->p[i]);

		T = last_write_2(ctx, dep, copy, j, read_level, &trest);
		rest = isl_set_union_disjoint(ctx, rest, trest);
		temp_rel[j] = isl_map_union_disjoint(ctx, temp_rel[j], T);
	    }
	    isl_set_free(ctx, todo);
	    todo = rest;

	    /* the middle write */
	    for (k = j-1; k >= 0; --k) {
		int plevel, plevel2;
		plevel = dep->precedes_level(dep->write[k].data, dep->read.data);
		if (plevel < read_level)
		    continue;
		if (!(read_level % 2) && plevel > read_level)
		    continue;

		plevel2 = dep->precedes_level(dep->write[j].data,
						dep->write[k].data);

		for (write_level = read_level; write_level < writelen; ++write_level) {
		    int i;
		    struct isl_map *T;

		    if (plevel2 < write_level)
			continue;
		    if (!(write_level % 2) && plevel2 > write_level)
			continue;

		    T = isl_map_empty(ctx, nparam, dep->write[j].n_input,
					dep->read.n_input);

		    /* the three write thing */
		    for (i = 0; i < temp_rel[j]->n; ++i) {
			struct isl_map *T2;
			struct isl_set *trest;
			struct isl_map *T_j;
			struct isl_basic_map *copy =
				isl_basic_map_copy(ctx, temp_rel[j]->p[i]);
			T2 = last_write_3(ctx, dep, copy, j, read_level, k, write_level,
					  &trest);
			T_j = isl_map_from_basic_map(ctx,
				isl_basic_map_copy(ctx, temp_rel[j]->p[i]));
			if (T2->n == 0) {
			    isl_set_free(ctx, trest);
			    isl_map_free(ctx, T2);
			    T = isl_map_union_disjoint(ctx, T, T_j);
			    continue;
			}
			T_j = isl_map_intersect_range(ctx, T_j, trest);
			T = isl_map_union_disjoint(ctx, T, T_j);
			temp_rel[k] = isl_map_union_disjoint(ctx, temp_rel[k],
								T2);
		    }
		    isl_map_free(ctx, temp_rel[j]);
		    temp_rel[j] = T;
		}
	    }

	    assert(temp_rel[j]);
	    final_rel[j].rel = Domain_concat(
				isl_map_to_polylib(ctx, temp_rel[j]),
				final_rel[j].rel);
	}

	for (j = dep->n_write-1; j >=0; --j)
	    isl_map_free(ctx, temp_rel[j]);
    }
    assert(todo);
    final_rel[dep->n_write].rel = isl_set_to_polylib(ctx, todo);
    free(temp_rel);
done:
    isl_set_free(ctx, todo);
    isl_ctx_free(ctx);
    return final_rel;
error:
    isl_ctx_free(ctx);
    return NULL;
}

static int copy_domain_constraints(Polyhedron *dep1, Polyhedron *dep2,
				   int n_write1, int n_write2,
				   int n_read, Polyhedron *C, 
				   PipMatrix *domain, int row, int col)
{
    int nparam = C->Dimension;
    int extra1 = dep1->Dimension - n_write1 - n_read - nparam;
    int extra2 = dep2->Dimension - n_write2 - n_read - nparam;

    poly2pipc(dep1, domain, n_write1, dep1->NbConstraints - n_write1, row, 
	       0, 1, 0);
    poly2pipc(dep1, domain, n_write1, dep1->NbConstraints - n_write1, row,
	      1+n_write1, n_read + extra1, 1+col);
    poly2pipc(dep1, domain, n_write1, dep1->NbConstraints - n_write1, row,
	      1+n_write1+n_read+extra1, nparam+1, domain->NbColumns-nparam-1);
    row += dep1->NbConstraints - n_write1;
    poly2pipc(dep2, domain, n_write2, dep2->NbConstraints - n_write2, row,
	       0, 1, 0);
    poly2pipc(dep2, domain, n_write2, dep2->NbConstraints - n_write2, row,
	      1+n_write2, n_read + extra2, 1+col+n_read+extra1);
    poly2pipc(dep2, domain, n_write2, dep2->NbConstraints - n_write2, row,
	      1+n_write2+n_read+extra2, nparam+1, domain->NbColumns-nparam-1);
    row += dep2->NbConstraints - n_write2;
    poly2pipc(C, domain, 0, C->NbConstraints, row, 0, 1, 0);
    poly2pipc(C, domain, 0, C->NbConstraints, row, 
	      1, nparam+1, domain->NbColumns-nparam-1);
    row += C->NbConstraints;
    return row;
}

/* only use on result of maximization problems */
static int lex_sign(PipList *list, int dim)
{
    int i;
    for (i = 0; i < dim; ++i) {
	assert(list);
	if (value_zero_p(list->vector->the_deno[0]))
	    return 1;
	if (value_neg_p(list->vector->the_vector[0]))
	    return -1;
	if (value_pos_p(list->vector->the_vector[0]))
	    return 1;
	list = list->next;
    }
    return 0;
}

/*
 * Check whether there is any reordering between dep1 and dep2.
 * Both are assumed to be "dependent" mappings, i.e., mappings
 * where the input (the write) is explicitly defined in terms
 * of the output (the read) through equations.
 * This code does not check for any "reordering" in the writes
 * corresponding to the same iteration in the reads of both dep1
 * and dep2.
 *
 * From the two dependences
 *
 *       write	read	extra	params+constant
 * dep1   I	 R1	 E1	 C1
 *	  0	 R3	 E3	 C3
 *
 * dep2   I	 R2	 E2	 C2
 *	  0	 R4	 E4	 C4
 * 
 * and the context
 *	                         C5
 *
 * we first construct
 *
 *	n_write n_read	    extra1  n_read  extra2  nparam+1
 *	z   	read1   	    read2	
 *	I   	R1	    E1      -R2	    -E2	    C1-C2   =>  z = write1 - write2
 *	0   	R3	    E3	    0	    0	    C3
 *	0   	0	    0	    R4	    E4	    C4
 *	0   	0	    0	    0	    0	    C5
 *
 * and add constraints that express that read1 is (strictly) lexicosmaller
 * than read2.
 *
 * The lexicomaximum of z is lexicopositive iff there exists any
 * write1 lexicolarger than write2.
 */
static int is_reordering_pair(Polyhedron *dep1, Polyhedron *dep2,
			      int n_write, int n_read, Polyhedron *C)
{
    PipOptions	*options;
    PipQuast	*sol;
    PipMatrix *domain, *ctx;
    PipList	*list;
    int nparam = C->Dimension;
    int ncol = 1 + dep1->Dimension + dep2->Dimension - n_write - nparam + 1;
    int nrow = dep1->NbConstraints + dep2->NbConstraints - n_write + n_read +
	       C->NbConstraints;
    int extra1 = dep1->Dimension - n_write - n_read - nparam;
    int extra2 = dep2->Dimension - n_write - n_read - nparam;
    int i, l;
    int	reordering = 0;
    int row = 0;
    ctx = pip_matrix_alloc(0, 2);

    assert_dependent_mapping(dep1, n_write);
    assert_dependent_mapping(dep2, n_write);

    options = pip_options_init();
    options->Simplify = 1;
    options->Maximize = 1;
    domain = pip_matrix_alloc(nrow, ncol);

    for (i = 0; i < n_write; ++i)
	value_set_si(domain->p[i][1+i], 1);
    poly2pipc(dep1, domain, 0, n_write, 0,
	      1+n_write, n_read + extra1, 1+n_write);
    poly2pipc(dep1, domain, 0, n_write, 0,
	      1+n_write+n_read+extra1, nparam+1, domain->NbColumns-nparam-1);
    sub_poly2pipc(dep2, domain, 0, n_write, 0,
		  1+n_write, n_read + extra2, 1+n_write+n_read+extra1);
    sub_poly2pipc(dep2, domain, 0, n_write, 0,
		  1+n_write+n_read+extra2, nparam+1, domain->NbColumns-nparam-1);
    row += n_write;
    row = copy_domain_constraints(dep1, dep2, n_write, n_write, n_read, 
				  C, domain, row, n_write);

    for (l = 0; l < n_read && !reordering; ++l) {
	var_equal(domain, 1+n_write, l, 1+n_write+n_read+extra1, row);
	domain->NbRows = row + l;
	var_smaller(domain, 1+n_write+l, 1+n_write+n_read+extra1+l, row+l);
	domain->NbRows++;

#ifdef PIP_DEBUG
	pip_matrix_print(stderr, domain);
	pip_matrix_print(stderr, ctx);
#endif

	sol = pip_solve(domain, ctx, -1, options);
#ifdef PIP_DEBUG
	pip_quast_print(stderr, sol, 0);
#endif

	if (sol->newparm || (sol->list && lex_sign(sol->list, n_write) > 0))
	    reordering = 1;

	pip_quast_free(sol);
    }
    pip_matrix_free(domain);
    pip_options_free(options);
    pip_matrix_free(ctx);
    return reordering;
}

/* For any two pairs of corresponding writes and reads,
 * do the writes occur in a different order than the reads ?
 * If so, then the dependence is "reordering".
 */
int pip_is_reordering(Polyhedron *dep, int n_write, int n_read, Polyhedron *ctx)
{
    Polyhedron *P1, *P2;
    for (P1 = dep; P1; P1 = P1->next)
	for (P2 = dep; P2; P2 = P2->next)
	    if (is_reordering_pair(P1, P2, n_write, n_read, ctx))
		return 1;
    return 0;
}

/*
 * Check whether the writes from dep1 and dep2 corresponding
 * to the same read iteration occur in order.
 * Both are assumed to be "dependent" mappings, i.e., mappings
 * where the input (the write) is explicitly defined in terms
 * of the output (the read) through equations.
 *
 * From the two dependences
 *
 *       write	read	extra	params+constant
 * dep1   I	 R1	 E1	 C1
 *	  0	 R3	 E3	 C3
 *
 * dep2   I	 R2	 E2	 C2
 *	  0	 R4	 E4	 C4
 * 
 * and the context
 *	                         C5
 *
 * we construct
 *
 *	n_write n_read	    extra1  extra2  nparam+1
 *	z   	read
 *	I   	R1-R2	    E1      -E2	    C1-C2   =>  z = write1 - write2
 *	0   	R3	    E3	    0	    C3
 *	0   	R4	    0	    E4	    C4
 *	0   	0	    0	    0	    C5
 *
 * The lexicomaximum of z is lexicopositive iff there exists any
 * write1 lexicolarger than write2.
 */
static int is_inorder_accesses_pair(Polyhedron *dep1, Polyhedron *dep2,
				    int n_write, int n_read, Polyhedron *C)
{
    PipOptions	*options;
    PipQuast	*sol;
    PipMatrix *domain, *ctx;
    PipList	*list;
    int nparam = C->Dimension;
    int ncol = 1 + dep1->Dimension + dep2->Dimension - n_write - n_read - nparam + 1;
    int nrow = dep1->NbConstraints + dep2->NbConstraints - n_write +
	       C->NbConstraints;
    int extra1 = dep1->Dimension - n_write - n_read - nparam;
    int extra2 = dep2->Dimension - n_write - n_read - nparam;
    int i, l;
    int	inorder = 1;
    int row = 0;
    ctx = pip_matrix_alloc(0, 2);

    assert_dependent_mapping(dep1, n_write);
    assert_dependent_mapping(dep2, n_write);

    options = pip_options_init();
    options->Simplify = 1;
    options->Maximize = 1;
    domain = pip_matrix_alloc(nrow, ncol);

    for (i = 0; i < n_write; ++i)
	value_set_si(domain->p[i][1+i], 1);
    poly2pipc(dep1, domain, 0, n_write, 0,
	      1+n_write, n_read + extra1, 1+n_write);
    poly2pipc(dep1, domain, 0, n_write, 0,
	      1+n_write+n_read+extra1, nparam+1, domain->NbColumns-nparam-1);
    sub_poly2pipc(dep2, domain, 0, n_write, 0,
		  1+n_write, n_read, 1+n_write);
    sub_poly2pipc(dep2, domain, 0, n_write, 0,
		  1+n_write+n_read, extra2, 1+n_write+n_read+extra1);
    sub_poly2pipc(dep2, domain, 0, n_write, 0,
		  1+n_write+n_read+extra2, nparam+1, domain->NbColumns-nparam-1);
    row += n_write;
    row = copy_domain_constraints(dep1, dep2, n_write, n_write, n_read, 
				  C, domain, row, n_write);

#ifdef PIP_DEBUG
    pip_matrix_print(stderr, domain);
    pip_matrix_print(stderr, ctx);
#endif

    sol = pip_solve(domain, ctx, -1, options);
#ifdef PIP_DEBUG
    pip_quast_print(stderr, sol, 0);
#endif

    if (sol->newparm || (sol->list && lex_sign(sol->list, n_write) > 0))
	inorder = 0;

    pip_quast_free(sol);
    pip_matrix_free(domain);
    pip_options_free(options);
    pip_matrix_free(ctx);
    return inorder;
}

/* For reads from dep1 and dep2 from the same iteration,
 * does the corresponding write from dep1 occur before the write from dep2 ?
 */
int pip_is_inorder_accesses(Polyhedron *dep1, Polyhedron *dep2,
			    int n_write, int n_read, Polyhedron *ctx)
{
    Polyhedron *P1, *P2;
    for (P1 = dep1; P1; P1 = P1->next)
	for (P2 = dep2; P2; P2 = P2->next)
	    if (!is_inorder_accesses_pair(P1, P2, n_write, n_read, ctx))
		return 0;
    return 1;
}

/* Check whether any read of dep1 occurs between a pair of write
 * and read of dep2
 *
 * From the two dependences
 *
 *       write	read	extra	params+constant
 * dep1   I	 R1	 E1	 C1
 *	  0	 R3	 E3	 C3
 *
 * dep2   I	 R2	 E2	 C2
 *	  0	 R4	 E4	 C4
 * 
 * and the context
 *	                         C5
 *
 * we first construct
 *
 *	n_write2    n_read  extra1  n_read  extra2  nparam+1
 *	z   	    read1   	    read2	
 *	I	    -I	    0	    R2	    E2	    C2	    =>  z = read1 - write2
 *	0   	    R3	    E3	    0	    0	    C3
 *	0   	    0	    0	    R4	    E4	    C4
 *	0   	    0	    0	    0	    0	    C5
 *
 * and add constraints that express that read1 is lexicosmaller than or equal to
 * (in case dep1 != dep2) read2.
 *
 * The lexicomaximum of z is lexicopositive iff there exists any
 * read1 lexicolarger than write2.
 */
static int is_consecutive_pair(Polyhedron *dep1, Polyhedron *dep2,
			       int n_write1, int n_read, Polyhedron *C)
{
    PipOptions	*options;
    PipQuast	*sol;
    PipMatrix *domain, *ctx;
    PipList	*list;
    int nparam = C->Dimension;
    int n_write2 = n_read;
    int ncol = 1 + dep1->Dimension + dep2->Dimension - n_write1 - nparam + 1;
    int nrow = dep1->NbConstraints + dep2->NbConstraints - n_write1 + n_read +
	       C->NbConstraints;
    int extra1 = dep1->Dimension - n_write1 - n_read - nparam;
    int extra2 = dep2->Dimension - n_write2 - n_read - nparam;
    int i, l;
    int	consecutive = 1;
    int row = 0;
    ctx = pip_matrix_alloc(0, 2);

    assert_dependent_mapping(dep1, n_write1);
    assert_dependent_mapping(dep2, n_write2);

    options = pip_options_init();
    options->Simplify = 1;
    options->Maximize = 1;
    domain = pip_matrix_alloc(nrow, ncol);

    assert(n_write2 == n_read);
    for (i = 0; i < n_write2; ++i) {
	value_set_si(domain->p[i][1+i], 1);
	value_set_si(domain->p[i][1+n_write2+i], -1);
    }
    sub_poly2pipc(dep2, domain, 0, n_write2, 0,
		  1+n_write2, n_read + extra2, 1+n_write2+n_read+extra1);
    sub_poly2pipc(dep2, domain, 0, n_write2, 0,
		  1+n_write2+n_read+extra2, nparam+1, domain->NbColumns-nparam-1);
    row += n_write2;
    row = copy_domain_constraints(dep1, dep2, n_write1, n_write2, n_read, 
				  C, domain, row, n_write2);

    for (l = 0; l <= n_read && consecutive; ++l) {
	if (l == n_read && dep1 == dep2)
	    break;
	var_equal(domain, 1+n_write2, l, 1+n_write2+n_read+extra1, row);
	domain->NbRows = row + l;
	if (l < n_read) {
	    var_smaller(domain, 1+n_write2+l, 1+n_write2+n_read+extra1+l, row+l);
	    domain->NbRows++;
	}

#ifdef PIP_DEBUG
	pip_matrix_print(stderr, domain);
	pip_matrix_print(stderr, ctx);
#endif

	sol = pip_solve(domain, ctx, -1, options);
#ifdef PIP_DEBUG
	pip_quast_print(stderr, sol, 0);
#endif

	if (sol->newparm || (sol->list && lex_sign(sol->list, n_write2) > 0))
	    consecutive = 0;

	pip_quast_free(sol);
    }
    pip_matrix_free(domain);
    pip_options_free(options);
    pip_matrix_free(ctx);
    return consecutive;
}

/* For any pair of corresponding write and read,
 * is there any other read in between ?
 * If so, then the loop is not "consecutive".
 * If two reads occur "simultaneously", we also do not consider 
 * the loop to be consecutive.
 */
int pip_is_consecutive_loop(Polyhedron *dep, int n_write, int n_read, 
			    Polyhedron *ctx)
{
    Polyhedron *P1, *P2;
    assert(n_write == n_read);
    for (P1 = dep; P1; P1 = P1->next)
	for (P2 = dep; P2; P2 = P2->next)
	    if (!is_consecutive_pair(P1, P2, n_write, n_read, ctx))
		return 0;
    return 1;
}

/* For any pair of corresponding write and read of dep2,
 * is there any read from dep1 in between ?
 * If so, then dep2 is not "consecutive".
 * If two reads occur "simultaneously", we also do not consider 
 * the loop to be consecutive.
 */
int pip_is_consecutive_dep(Polyhedron *dep2, Polyhedron *dep1, 
			   int n_write, int n_read, Polyhedron *ctx)
{
    Polyhedron *P1, *P2;
    for (P1 = dep1; P1; P1 = P1->next)
	for (P2 = dep2; P2; P2 = P2->next)
	    if (!is_consecutive_pair(P1, P2, n_write, n_read, ctx))
		return 0;
    return 1;
}

/*
 * Check whether there is a pair of distinct reads, one from dep1 and dep2,
 * that correspond to the same write.
 * Both dependences are assumed to be "dependent" mappings, i.e., mappings
 * where the input (the write) is explicitly defined in terms
 * of the output (the read) through equations.
 *
 * From the two dependences
 *
 *       write	read	extra	params+constant
 * dep1   I	 R1	 E1	 C1
 *	  0	 R3	 E3	 C3
 *
 * dep2   I	 R2	 E2	 C2
 *	  0	 R4	 E4	 C4
 * 
 * and the context
 *	                         C5
 *
 * we construct
 *
 *	n_read 	n_read	    extra1  n_read  extra2  nparam+1
 *	z   	read1   	    read2	
 *	I   	-I	    0	    I	    0	    0		=>  z = read1 - read2
 *	0   	R1	    E1	    -R2	    -E2	    C1-C2	=>  write1 = write2
 *	0   	R3	    E3	    0	    0	    C3
 *	0   	0	    0	    R4	    E4	    C4
 *	0   	0	    0	    0	    0	    C5
 *
 * If the lexicomaximal z is lexicopositive then such distinct reads exist.
 * Note that if the lexicomaximal z is lexiconegative, then such reads also
 * exist, but we choose to return false and this case.  Calling the function
 * again with dep1 and dep2 interchanged will return true.
 * If the lexicomaximal z is zero and if dep1 != dep2, then such reads
 * _may_ exist.  Again we return false in this case since calling the function
 * with dep1 and dep2 interchanged will determine whether they exist.
 */
static int has_multiplicity_pair(Polyhedron *dep1, Polyhedron *dep2,
			         int n_write, int n_read, Polyhedron *C)
{
    PipOptions	*options;
    PipQuast	*sol;
    PipMatrix *domain, *ctx;
    PipList	*list;
    int nparam = C->Dimension;
    int nrow = dep1->NbConstraints + dep2->NbConstraints - n_write + n_read +
	       C->NbConstraints;
    int extra1 = dep1->Dimension - n_write - n_read - nparam;
    int extra2 = dep2->Dimension - n_write - n_read - nparam;
    int ncol = 1 + n_read + n_read + extra1 + n_read + extra2 + nparam + 1;
    int i;
    int	multiplicity = 0;
    int row = 0;
    ctx = pip_matrix_alloc(0, 2);

    assert_dependent_mapping(dep1, n_write);
    assert_dependent_mapping(dep2, n_write);

    options = pip_options_init();
    options->Simplify = 1;
    options->Maximize = 1;
    domain = pip_matrix_alloc(nrow, ncol);

    for (i = 0; i < n_read; ++i) {
	value_set_si(domain->p[i][1+i], 1);
	value_set_si(domain->p[i][1+n_read+i], -1);
	value_set_si(domain->p[i][1+n_read+n_read+extra1+i], 1);
    }
    row += n_read;
    poly2pipc(dep1, domain, 0, n_write, row,
	      1+n_write, n_read + extra1, 1+n_read);
    poly2pipc(dep1, domain, 0, n_write, row,
	      1+n_write+n_read+extra1, nparam+1, domain->NbColumns-nparam-1);
    sub_poly2pipc(dep2, domain, 0, n_write, row,
		  1+n_write, n_read + extra2, 1+n_read+n_read+extra1);
    sub_poly2pipc(dep2, domain, 0, n_write, row,
		  1+n_write+n_read+extra2, nparam+1, domain->NbColumns-nparam-1);
    row += n_write;
    row = copy_domain_constraints(dep1, dep2, n_write, n_write, n_read, 
				  C, domain, row, n_read);

#ifdef PIP_DEBUG
    pip_matrix_print(stderr, domain);
    pip_matrix_print(stderr, ctx);
#endif

    sol = pip_solve(domain, ctx, -1, options);
#ifdef PIP_DEBUG
    pip_quast_print(stderr, sol, 0);
#endif

    if (sol->newparm || (sol->list && lex_sign(sol->list, n_read) > 0))
	multiplicity = 1;

    pip_quast_free(sol);
    pip_matrix_free(domain);
    pip_options_free(options);
    pip_matrix_free(ctx);
    return multiplicity;
}

int pip_has_multiplicity(Polyhedron *dep, int n_write, int n_read, Polyhedron *ctx)
{
    Polyhedron *P1, *P2;
    for (P1 = dep; P1; P1 = P1->next)
	for (P2 = dep; P2; P2 = P2->next)
	    if (has_multiplicity_pair(P1, P2, n_write, n_read, ctx))
		return 1;
    return 0;
}

/*
 * Compute the minimal distance vector for dependence "dep", with
 * the writes mapped to common iteration space of dimension "dim"
 * through "sw" and the reads through "sr".
 * Both are assumed to be "dependent" mappings, i.e., mappings
 * where the input (the iteration vector in the common iteration
 * space) is explicitly defined in terms of the output (the iteration
 * vector in the original space) through equations.
 *
 * From
 *	dim	read	sr_extra	params+constant
 * sr	 I	 R1	 E1		 C1		    => r' = -R1 r ...
 *
 *	dim	write	sw_extra	params+constant
 * sw	 I	 W1	 E2		 C2		    => w' = -W1 w ...
 *
 *	write	read	d_extra		params+constant
 * dep	 W2	 R2	 E3		 C3
 *
 * we first construct
 *
 *  d   dim	write   sw_extra    read    sr_extra    d_extra	params+constant
 *	 I       -W1	 -E1	     R1	     E2			 C2-C1
 *		  W2		     R2			  E3	 C3
 *  => d = r'-w'
 *
 * and compute the lexicographically minimal of these distance vectors.
 */
Polyhedron *pip_min_ddv(Polyhedron *dep, Polyhedron *sw, Polyhedron *sr,
			int nwrite, int nread, int dim, Polyhedron *C)
{
    Polyhedron *R;
    int nparam = C->Dimension;
    struct isl_ctx *ctx;
    struct isl_basic_map *dep_map;
    struct isl_basic_map *sw_map;
    struct isl_basic_map *sr_map;
    struct isl_basic_set *ddv;
    struct isl_set *min;

    ctx = isl_ctx_alloc();
    dep_map = isl_basic_map_new_from_polylib(ctx, dep,
						 nparam, nwrite, nread);
    sw_map = isl_basic_map_new_from_polylib(ctx, sw, nparam, dim, nwrite);
    sr_map = isl_basic_map_new_from_polylib(ctx, sr, nparam, dim, nread);
    sw_map = isl_basic_map_reverse(ctx, sw_map);
    sr_map = isl_basic_map_reverse(ctx, sr_map);
    dep_map = isl_basic_map_apply_domain(ctx, dep_map, sw_map);
    dep_map = isl_basic_map_apply_range(ctx, dep_map, sr_map);

    ddv = isl_basic_map_deltas(ctx, dep_map);
    min = isl_basic_set_lexmin(ctx, ddv);
    R = isl_set_to_polylib(ctx, min);
    isl_set_free(ctx, min);

    isl_ctx_free(ctx);

    return R;
}
